"""
Filter rows by subject, predicate, object values.
"""
from argparse import Namespace, SUPPRESS
import typing

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

def parser():
    return {
        'help': 'Filter rows by subject, predicate, object values.',
        'description': 'Filter KGTK file based on values in the node1 (subject), ' +
        'label (predicate), and node2 (object) fields.  Optionally filter based on ' +
        'regular expressions.'
    }


REGEX_MATCH_TYPES: typing.List[str] = [ "fullmatch", "match","search" ]
NUMERIC_MATCH_TYPES: typing.List[str] = [ "eq", "ne", "gt", "ge", "lt", "le" ]
MATCH_TYPES: typing.List[str] = REGEX_MATCH_TYPES + NUMERIC_MATCH_TYPES

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions, KgtkReaderMode
    from kgtk.utils.argparsehelpers import optional_bool
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    _expert: bool = parsed_shared_args._expert

    # '$label == "/r/DefinedAs" && $node2=="/c/en/number_zero"'
    parser.add_input_file(positional=True)
    parser.add_output_file(who="The KGTK output file for records that pass the filter. Multiple output file may be specified, each with their own pattern.", allow_list=True, dest="output_files")
    parser.add_output_file(who="The KGTK reject file for records that fail the filter.",
                           dest="reject_file",
                           options=["--reject-file"],
                           metavar="REJECT_FILE",
                           optional=True)

    # parser.add_argument('-dt', "--datatype", action="store", type=str, dest="datatype", help="Datatype of the input file, e.g., tsv or csv.", default="tsv")
    parser.add_argument('-p', '--pattern', action="append", nargs="+", type=str, dest="patterns", required=True,
                        help="Pattern to filter on, for instance, \" ; P154 ; \". Multiple patterns may be specified when there are mutiple output files.")
    parser.add_argument('--node1', '--subj', action="store", type=str, dest='subj_col', help="The subject column, default is node1 or its alias.")
    parser.add_argument('--label', '--pred', action="store", type=str, dest='pred_col', help="The predicate column, default is label or its alias.")
    parser.add_argument('--node2', '--obj', action="store", type=str, dest='obj_col', help="The object column, default is node2 or its alias.")

    parser.add_argument(      "--or", dest="or_pattern", metavar="True|False",
                              help="'Or' the clauses of the pattern. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--invert", dest="invert", metavar="True|False",
                              help="Invert the result of applying the pattern. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--regex", dest="regex", metavar="True|False",
                              help="When True, treat the filter clauses as regular expressions. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--numeric", dest="numeric", metavar="True|False",
                              help="When True, treat the filter clauses as numeric values for comparison. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--fancy", dest="fancy", metavar="True|False",
                              help="When True, treat the filter clauses as strings, numbers, or regular expressions. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--match-type", dest="match_type", 
                              help="Which type of regular expression match: %(choices)s. (default=%(default)s).",
                              type=str, default="match", choices=MATCH_TYPES)

    parser.add_argument(      "--first-match-only", dest="first_match_only", metavar="True|False",
                              help="If true, write only to the file with the first matching pattern.  If false, write to all files with matching patterns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument(      "--pass-empty-value", dest="pass_empty_value", metavar="True|False",
                              help="If true, empty data values will pass a numeric pattern.  If false, write to all files with matching patterns. (default=%(default)s).",
                              type=optional_bool, nargs='?', const=True, default=False)

    parser.add_argument('--pattern-separator', action="store", type=str, dest='pattern_separator', default=";",
                        help="The separator between the pattern components. (default=%(default)s.")

    parser.add_argument('--word-separator', action="store", type=str, dest='word_separator', default=",",
                        help="The separator between the words in a pattern component. (default=%(default)s.")

    parser.add_argument(      "--show-version", dest="show_version", type=optional_bool, nargs='?', const=True, default=False,
                              help="Print the version of this program. (default=%(default)s).", metavar="True/False")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser,
                                    mode_options=True,
                                    default_mode=KgtkReaderMode[parsed_shared_args._mode],
                                    expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)

def run(input_file: KGTKFiles,
        output_files: KGTKFiles,
        reject_file: KGTKFiles,

        patterns: typing.List[typing.List[str]],
        subj_col: typing.Optional[str],
        pred_col: typing.Optional[str],
        obj_col: typing.Optional[str],

        or_pattern: bool,
        invert: bool,
        regex: bool,
        numeric: bool,
        fancy: bool,
        match_type: str,
        first_match_only: bool,
        pass_empty_value: bool,

        pattern_separator: str = ";",
        word_separator: str = ",",

        show_version: bool = False,

        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,

        **kwargs # Whatever KgtkFileOptions and KgtkValueOptions want.
)->int:
    # import modules locally
    from pathlib import Path
    import re
    import sys
    
    from kgtk.exceptions import kgtk_exception_auto_handler, KGTKException
    from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
    from kgtk.io.kgtkwriter import KgtkWriter
    from kgtk.value.kgtkvalueoptions import KgtkValueOptions

    input_kgtk_file: Path = KGTKArgumentParser.get_input_file(input_file)
    output_kgtk_files: typing.List[Path] = KGTKArgumentParser.get_output_file_list(output_files, default_stdout=True)
    reject_kgtk_file: typing.Optional[Path] = KGTKArgumentParser.get_optional_output_file(reject_file, who="KGTK reject file")

    # Select where to send error messages, defaulting to stderr.
    error_file: typing.TextIO = sys.stdout if errors_to_stdout else sys.stderr

    # Build the option structures.
    reader_options: KgtkReaderOptions = KgtkReaderOptions.from_dict(kwargs)
    value_options: KgtkValueOptions = KgtkValueOptions.from_dict(kwargs)

    UPDATE_VERSION: str = "2021-09-24T02:35:27.840163+00:00#gysblgql6Q7482L14Zozt/ne8Owd497FJa7MVp92+UbmixJKElkfg/GY5UmGBsog86NPtmYy+dXWa6PRMIyuIw=="
    if show_version or verbose:
        print("kgtk filter version: %s" % UPDATE_VERSION, file=error_file, flush=True)

    # Show the final option structures for debugging and documentation.
    if show_options:
        print("--input-file=%s" % str(input_kgtk_file), file=error_file)
        print("--output-file=%s" % " ".join([str(x) for x in output_kgtk_files]), file=error_file)
        if reject_kgtk_file is not None:
            print("--reject-file=%s" % str(reject_kgtk_file), file=error_file)
        print("--pattern=%s" % " ".join(repr(patterns)), file=error_file)
        if subj_col is not None:
            print("--subj=%s" % str(subj_col), file=error_file)
        if pred_col is not None:
            print("--pred=%s" % str(pred_col), file=error_file)
        if obj_col is not None:
            print("--obj=%s" % str(obj_col), file=error_file)
        print("--or=%s" % str(or_pattern), file=error_file)
        print("--invert=%s" % str(invert), file=error_file)
        print("--regex=%s" % str(regex), file=error_file)
        print("--numeric=%s" % str(numeric), file=error_file)
        print("--fancy=%s" % str(fancy), file=error_file)
        print("--match-type=%s" % repr(match_type), file=error_file)
        print("--first-match-only=%s" % str(first_match_only), file=error_file)
        print("--pass_empty_value=%s" % str(pass_empty_value), file=error_file)
        print("--pattern-separator %s" % repr(pattern_separator), file=error_file)
        print("--word-separator %s" % repr(word_separator), file=error_file)
        reader_options.show(out=error_file)
        value_options.show(out=error_file)
        print("=======", file=error_file, flush=True)


    if regex and numeric:
        raise KGTKException("Select either --regex or --numeric, not both")

    if regex and fancy:
        raise KGTKException("Select either --regex or --fancy, not both")

    if numeric and fancy:
        raise KGTKException("Select either --numeric or --fancy, not both")

    if regex and match_type not in REGEX_MATCH_TYPES:
        raise KGTKException("Match type %s in not a regex match type." % repr(match_type))

    if numeric and match_type not in NUMERIC_MATCH_TYPES:
        raise KGTKException("Match type %s in not a numeric match type." % repr(match_type))

    if fancy and match_type not in REGEX_MATCH_TYPES:
        raise KGTKException("Match type %s in not a regex match type for fancy mode." % repr(match_type))

    def prepare_filter(pattern: str)->typing.Set[str]:
        filt: typing.Set[str] = set()
        pattern = pattern.strip()
        if len(pattern) == 0:
            return filt

        target: str
        for target in pattern.split(word_separator):
            target=target.strip()
            if len(target) > 0:
                filt.add(target)

        return filt

    def prepare_regex(pattern: str)->typing.Optional[typing.Pattern]:
        pattern = pattern.strip()
        if len(pattern) == 0:
            return None
        else:
            return re.compile(pattern)

    def single_subject_filter(kr: KgtkReader,
                              kw: KgtkWriter,
                              rw: typing.Optional[KgtkWriter],
                              subj_idx: int,
                              subj_filter: typing.Set[str],
                              ):
        if verbose:
            print("Applying a single subject filter", file=error_file, flush=True)

        subj_filter_value: str = list(subj_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[subj_idx] == subj_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_subject_filter_inverted(kr: KgtkReader,
                                       kw: KgtkWriter,
                                       rw: typing.Optional[KgtkWriter],
                                       subj_idx: int,
                                       subj_filter: typing.Set[str],
                                       ):
        if verbose:
            print("Applying a single subject filter inverted", file=error_file, flush=True)

        subj_filter_value: str = list(subj_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[subj_idx] != subj_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_predicate_filter(kr: KgtkReader,
                                kw: KgtkWriter,
                                rw: typing.Optional[KgtkWriter],
                                pred_idx: int,
                                pred_filter: typing.Set[str],
                                ):
        if verbose:
            print("Applying a single predicate filter", file=error_file, flush=True)

        pred_filter_value: str = list(pred_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[pred_idx] == pred_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_predicate_filter_inverted(kr: KgtkReader,
                                         kw: KgtkWriter,
                                         rw: typing.Optional[KgtkWriter],
                                         pred_idx: int,
                                         pred_filter: typing.Set[str],
                                         ):
        if verbose:
            print("Applying a single predicate filter inverted", file=error_file, flush=True)

        pred_filter_value: str = list(pred_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[pred_idx] != pred_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_object_filter(kr: KgtkReader,
                             kw: KgtkWriter,
                             rw: typing.Optional[KgtkWriter],
                             obj_idx: int,
                             obj_filter: typing.Set[str],
                             ):
        if verbose:
            print("Applying a single object filter", file=error_file, flush=True)

        obj_filter_value: str = list(obj_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[obj_idx] == obj_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_object_filter_inverted(kr: KgtkReader,
                                      kw: KgtkWriter,
                                      rw: typing.Optional[KgtkWriter],
                                      obj_idx: int,
                                      obj_filter: typing.Set[str],
                                      ):
        if verbose:
            print("Applying a single object filter inverted", file=error_file, flush=True)

        obj_filter_value: str = list(obj_filter)[0]

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            if row[obj_idx] != obj_filter_value:
                kw.write(row)
                output_line_count += 1

            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def single_general_filter(kr: KgtkReader,
                              kw: KgtkWriter,
                              rw: typing.Optional[KgtkWriter],
                              subj_idx: int,
                              subj_filter: typing.Set[str],
                              pred_idx: int,
                              pred_filter: typing.Set[str],
                              obj_idx: int,
                              obj_filter: typing.Set[str]):
        if verbose:
            print("Applying a single general filter", file=error_file, flush=True)

        apply_subj_filter: bool = len(subj_filter) > 0
        apply_pred_filter: bool = len(pred_filter) > 0
        apply_obj_filter: bool = len(obj_filter) > 0

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            keep: bool = False
            reject: bool = False 
            if apply_subj_filter:
                if row[subj_idx] in subj_filter:
                    keep = True
                    subj_filter_keep_count += 1
                else:
                    reject = True
                    subj_filter_reject_count += 1

            if apply_pred_filter:
                if row[pred_idx] in pred_filter:
                    keep = True
                    pred_filter_keep_count += 1
                else:
                    reject = True
                    pred_filter_reject_count += 1

            if apply_obj_filter:
                if row[obj_idx] in obj_filter:
                    keep = True
                    obj_filter_keep_count += 1
                else:
                    reject = True
                    obj_filter_reject_count += 1

            if (keep if or_pattern else not reject) ^ invert:
                kw.write(row)
                output_line_count += 1
            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def dispatch_subject_filter(kr: KgtkReader,
                                kws: typing.List[KgtkWriter],
                                rw: typing.Optional[KgtkWriter],
                                subj_idx: int,
                                subj_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a dispatched multiple-output subject filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        dispatch: typing.MutableMapping[str, KgtkWriter] = { }
        idx: int
        kw: KgtkWriter
        for idx, kw in enumerate(kws):
            subj_filter: typing.Set[str] = subj_filters[idx]
            keyword: str
            for keyword in subj_filter:
                dispatch[keyword] = kw

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            kwo: typing.Optional[KgtkWriter] = dispatch.get(row[subj_idx])
            if kwo is not None:
                kwo.write(row)
                output_line_count += 1
            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def dispatch_predicate_filter(kr: KgtkReader,
                                  kws: typing.List[KgtkWriter],
                                  rw: typing.Optional[KgtkWriter],
                                  pred_idx: int,
                                  pred_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a dispatched multiple-output predicate filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        dispatch: typing.MutableMapping[str, KgtkWriter] = { }
        idx: int
        kw: KgtkWriter
        for idx, kw in enumerate(kws):
            pred_filter: typing.Set[str] = pred_filters[idx]
            keyword: str
            for keyword in pred_filter:
                dispatch[keyword] = kw

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            kwo: typing.Optional[KgtkWriter] = dispatch.get(row[pred_idx])
            if kwo is not None:
                kwo.write(row)
                output_line_count += 1
            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def dispatch_object_filter(kr: KgtkReader,
                               kws: typing.List[KgtkWriter],
                               rw: typing.Optional[KgtkWriter],
                               obj_idx: int,
                               obj_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a dispatched multiple-output object filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0

        dispatch: typing.MutableMapping[str, KgtkWriter] = { }
        idx: int
        kw: KgtkWriter
        for idx, kw in enumerate(kws):
            obj_filter: typing.Set[str] = obj_filters[idx]
            keyword: str
            for keyword in obj_filter:
                dispatch[keyword] = kw

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            kwo: typing.Optional[KgtkWriter] = dispatch.get(row[obj_idx])
            if kwo is not None:
                kwo.write(row)
                output_line_count += 1
            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)

    def multiple_general_filter(kr: KgtkReader,
                                kws: typing.List[KgtkWriter],
                                rw: typing.Optional[KgtkWriter],
                                subj_idx: int,
                                subj_filters: typing.List[typing.Set[str]],
                                pred_idx: int,
                                pred_filters: typing.List[typing.Set[str]],
                                obj_idx: int,
                                obj_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a multiple-output general filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Set[str] = subj_filters[idx]
                pred_filter: typing.Set[str] = pred_filters[idx]
                obj_filter: typing.Set[str] = obj_filters[idx]
                idx += 1
                

                keep: bool = False
                reject: bool = False 
                if len(subj_filter) > 0:
                    if row[subj_idx] in subj_filter:
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if len(pred_filter) > 0:
                    if row[pred_idx] in pred_filter:
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if len(obj_filter) > 0:
                    if row[obj_idx] in obj_filter:
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def numeric_filter_test(value: str, threshold: float)->bool:
        if len(value) == 0:
            return pass_empty_value

        # TODO: catch a conversion failure and provide better feedback
        # and other options.
        try:
            numeric_value: float = float(value)
        except ValueError:
            return False

        if match_type == "eq":
            return numeric_value == threshold
        elif match_type == "ne":
            return numeric_value != threshold
        elif match_type == "gt":
            return numeric_value > threshold
        elif match_type == "ge":
            return numeric_value >= threshold
        elif match_type == "lt":
            return numeric_value < threshold
        elif match_type == "le":
            return numeric_value <= threshold
        else:
            raise KGTKException("Unknown match type %s during numeric filter test" % repr(match_type))

    def multiple_general_numeric_filter(kr: KgtkReader,
                                        kws: typing.List[KgtkWriter],
                                        rw: typing.Optional[KgtkWriter],
                                        subj_idx: int,
                                        subj_filters: typing.List[typing.Set[str]],
                                        pred_idx: int,
                                        pred_filters: typing.List[typing.Set[str]],
                                        obj_idx: int,
                                        obj_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a multiple general numeric filter", file=error_file, flush=True)

        # Convert the string-based filters to numeric filters.
        filter_set: typing.Set[str]
        subj_numeric_filters: typing.List[typing.Optional[float]] = [ ]
        for filter_set in subj_filters:
            if len(filter_set) > 1:
                raise KGTKException ("Numeric subject filters must be singleton values.")
            # TODO: Catch a conversion failure and give better feedback.
            if len(filter_set) > 0:
                subj_numeric_filters.append(float(list(filter_set)[0]))
            else:
                subj_numeric_filters.append(None)
            
        pred_numeric_filters: typing.List[typing.Optional[float]] = [ ]
        for filter_set in pred_filters:
            if len(filter_set) > 1:
                raise KGTKException ("Numeric predicate filters must be singleton values.")
            if len(filter_set) > 0:
                # TODO: Catch a conversion failure and give better feedback.
                pred_numeric_filters.append(float(list(filter_set)[0]))
            else:
                pred_numeric_filters.append(None)
            
        obj_numeric_filters: typing.List[typing.Optional[float]] = [ ]
        for filter_set in obj_filters:
            if len(filter_set) > 1:
                raise KGTKException ("Numeric object filters must be singleton values.")
            if len(filter_set) > 0:
                # TODO: Catch a conversion failure and give better feedback.
                obj_numeric_filters.append(float(list(filter_set)[0]))
            else:
                obj_numeric_filters.append(None)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Optional[float] = subj_numeric_filters[idx]
                pred_filter: typing.Optional[float] = pred_numeric_filters[idx]
                obj_filter: typing.Optional[float] = obj_numeric_filters[idx]
                idx += 1

                keep: bool = False
                reject: bool = False 
                if subj_filter is not None:
                    if numeric_filter_test(row[subj_idx], subj_filter):
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if pred_filter is not None:
                    if numeric_filter_test(row[pred_idx], pred_filter):
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if obj_filter is not None:
                    if numeric_filter_test(row[obj_idx], obj_filter):
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def single_general_numeric_filter(kr: KgtkReader,
                                      kws: typing.List[KgtkWriter],
                                      rw: typing.Optional[KgtkWriter],
                                      subj_idx: int,
                                      subj_filter: typing.Set[str],
                                      pred_idx: int,
                                      pred_filter: typing.Set[str],
                                      obj_idx: int,
                                      obj_filter: typing.Set[str]):
        if verbose:
            print("Applying a single general numeric filter", file=error_file, flush=True)

        # Convert the string-based filter to a numeric filter.
        subj_numeric_filter: typing.Optional[float]
        if len(subj_filter) > 1:
            raise KGTKException ("Numeric subject filters must be singleton values.")
        # TODO: Catch a conversion failure and give better feedback.
        elif len(subj_filter) == 1:
            subj_numeric_filter = float(list(subj_filter)[0])
        else:
            subj_numeric_filter = None
            
        pred_numeric_filter: typing.Optional[float]
        if len(pred_filter) > 1:
            raise KGTKException ("Numeric predicate filters must be singleton values.")
        elif len(pred_filter) == 1:
            # TODO: Catch a conversion failure and give better feedback.
            pred_numeric_filter = float(list(pred_filter)[0])
        else:
            pred_numeric_filter = None
            
        obj_numeric_filter: typing.Optional[float]
        if len(obj_filter) > 1:
            raise KGTKException ("Numeric object filters must be singleton values.")
        elif len(obj_filter) == 1:
            # TODO: Catch a conversion failure and give better feedback.
            obj_numeric_filter = float(list(obj_filter)[0])
        else:
            obj_numeric_filter = None

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        kw: KgtkWriter = kws[0]

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            keep: bool = False
            reject: bool = False 
            if subj_numeric_filter is not None:
                if numeric_filter_test(row[subj_idx], subj_numeric_filter):
                    keep = True
                    subj_filter_keep_count += 1
                else:
                    reject = True
                    subj_filter_reject_count += 1

            if pred_numeric_filter is not None:
                if numeric_filter_test(row[pred_idx], pred_numeric_filter):
                    keep = True
                    pred_filter_keep_count += 1
                else:
                    reject = True
                    pred_filter_reject_count += 1

            if obj_numeric_filter is not None:
                if numeric_filter_test(row[obj_idx], obj_numeric_filter):
                    keep = True
                    obj_filter_keep_count += 1
                else:
                    reject = True
                    obj_filter_reject_count += 1

            if (keep if or_pattern else not reject) ^ invert:
                kw.write(row)
                output_line_count += 1
            else:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def fancy_filter_test(value: str, fancy_filter: typing.Set[str])->bool:
        
        fancy_pattern: str
        for fancy_pattern in fancy_filter:
            # TODO: Optimize so the patterns aren't converted to float every time they're used.

            if fancy_pattern.startswith(":"):
                if value == fancy_pattern[1:]:
                    return True

            elif fancy_pattern.startswith("="):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) == float(fancy_pattern[1:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith("!="):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) != float(fancy_pattern[2:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith(">="):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) >= float(fancy_pattern[2:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith(">"):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) > float(fancy_pattern[1:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith("<="):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) <= float(fancy_pattern[2:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith("<"):
                if len(value) == 0 and pass_empty_value:
                    return True
                try:
                    if float(value) < float(fancy_pattern[1:]):
                        return True
                except ValueError:
                    pass

            elif fancy_pattern.startswith("~"):
                # TODO: optimize thie code so the regular expressions are not
                # compiled each time they're used.
                if match_type == "fullmatch":
                    if re.compile(fancy_pattern[1:]).fullmatch(value):
                        return True
                elif match_type == "match":
                    if re.compile(fancy_pattern[1:]).match(value):
                        return True
                elif match_type == "seach":
                    if re.compile(fancy_pattern[1:]).search(value):
                        return True
                else:
                    raise KGTKException("Match type %s is not valid for fancy filters." % repr(match_type))

            else:
                raise KGTKException("Unknown prefix in fancy pattern %s" % repr(fancy_pattern))
            

        return False

    def multiple_general_fancy_filter(kr: KgtkReader,
                                      kws: typing.List[KgtkWriter],
                                      rw: typing.Optional[KgtkWriter],
                                      subj_idx: int,
                                      subj_filters: typing.List[typing.Set[str]],
                                      pred_idx: int,
                                      pred_filters: typing.List[typing.Set[str]],
                                      obj_idx: int,
                                      obj_filters: typing.List[typing.Set[str]]):
        if verbose:
            print("Applying a multiple general fancy filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Set[str] = subj_filters[idx]
                pred_filter: typing.Set[str] = pred_filters[idx]
                obj_filter: typing.Set[str] = obj_filters[idx]
                idx += 1

                keep: bool = False
                reject: bool = False 
                if len(subj_filter) > 0:
                    if fancy_filter_test(row[subj_idx], subj_filter):
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if len(pred_filter) > 0:
                    if fancy_filter_test(row[pred_idx], pred_filter):
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if len(obj_filter) > 0:
                    if fancy_filter_test(row[obj_idx], obj_filter):
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def process_plain()->int:

        subj_filters: typing.List[typing.Set[str]] = [ ]
        pred_filters: typing.List[typing.Set[str]] = [ ]
        obj_filters: typing.List[typing.Set[str]] = [ ]

        subj_filter: typing.Set[str]
        pred_filter: typing.Set[str]
        obj_filter: typing.Set[str]


        nfilters: int = 0
        pattern_list: typing.List[str]
        pattern: str
        for pattern_list in patterns:
            for pattern in pattern_list:
                subpatterns: typing.List[str] = pattern.split(pattern_separator)
                if len(subpatterns) != 3:
                    print("Error: The pattern must have three sections separated by %s (two %s total)." % (repr(pattern_separator), repr(pattern_separator)),
                          file=error_file, flush=True)
                    raise KGTKException("Bad pattern")
            
                subj_filter = prepare_filter(subpatterns[0])
                pred_filter = prepare_filter(subpatterns[1])
                obj_filter = prepare_filter(subpatterns[2])

                if len(subj_filter) == 0 and len(pred_filter) == 0 and len(obj_filter) == 0:
                    if verbose:
                        print("Warning: the filter %s is empty." % repr(pattern), file=error_file, flush=True)
                else:
                    subj_filters.append(subj_filter)
                    pred_filters.append(pred_filter)
                    obj_filters.append(obj_filter)
                    nfilters += 1

        if nfilters == 0:
            raise KGTKException("No filters found.")

        if nfilters != len(output_kgtk_files):
            if verbose:
                print("output files: %s" % " ".join([str(x) for x in output_kgtk_files]), file=error_file, flush=True)
            raise KGTKException("There were %d filters and %d output files." % (nfilters, len(output_kgtk_files)))

        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        subj_idx: int = kr.get_node1_column_index(subj_col)
        pred_idx: int = kr.get_label_column_index(pred_col)
        obj_idx: int = kr.get_node2_column_index(obj_col)

        # Complain about a missing column only when it is needed by the pattern.
        trouble: bool = False
        if subj_idx < 0 and len(set.union(*subj_filters)) > 0:
            trouble = True
            print("Error: Cannot find the subject column '%s'." % kr.get_node1_canonical_name(subj_col), file=error_file, flush=True)
        if pred_idx < 0 and len(set.union(*pred_filters)) > 0:
            trouble = True
            print("Error: Cannot find the predicate column '%s'." % kr.get_label_canonical_name(pred_col), file=error_file, flush=True)
        if obj_idx < 0 and len(set.union(*obj_filters)) > 0:
            trouble = True
            print("Error: Cannot find the object column '%s'." % kr.get_node2_canonical_name(obj_col), file=error_file, flush=True)
        if trouble:
            # Clean up:
            kr.close()
            raise KGTKException("Missing columns.")

        kw: KgtkWriter
        kws: typing.List[KgtkWriter] = [ ]
        output_kgtk_file: Path
        for output_kgtk_file in output_kgtk_files:
            if verbose:
                print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
            kw = KgtkWriter.open(kr.column_names,
                                 output_kgtk_file,
                                 mode=KgtkWriter.Mode[kr.mode.name],
                                 use_mgzip=reader_options.use_mgzip, # Hack!
                                 mgzip_threads=reader_options.mgzip_threads, # Hack!
                                 error_file=error_file,
                                 verbose=verbose,
                                 very_verbose=very_verbose)
            kws.append(kw)

        rw: typing.Optional[KgtkWriter] = None
        if reject_kgtk_file is not None:
            if verbose:
                print("Opening the reject file: %s" % str(reject_kgtk_file), file=error_file, flush=True)
            rw = KgtkWriter.open(kr.column_names,
                                 reject_kgtk_file,
                                 mode=KgtkWriter.Mode[kr.mode.name],
                                 use_mgzip=reader_options.use_mgzip, # Hack!
                                 mgzip_threads=reader_options.mgzip_threads, # Hack!
                                 error_file=error_file,
                                 verbose=verbose,
                                 very_verbose=very_verbose)

        try:
            if fancy:
                # TODO: add additional optimized cases.
                multiple_general_fancy_filter(kr, kws, rw, subj_idx, subj_filters, pred_idx, pred_filters, obj_idx, obj_filters)

            elif numeric:
                # TODO: add additional optimized cases.
                
                if nfilters == 1:
                    single_general_numeric_filter(kr, kws, rw, subj_idx, subj_filters[0], pred_idx, pred_filters[0], obj_idx, obj_filters[0])
                else:
                    multiple_general_numeric_filter(kr, kws, rw, subj_idx, subj_filters, pred_idx, pred_filters, obj_idx, obj_filters)
                
            elif nfilters == 1:
                subj_filter = subj_filters[0]
                pred_filter = pred_filters[0]
                obj_filter = obj_filters[0]
                kw = kws[0]

                if len(subj_filter) == 1 and len(pred_filter) == 0 and len(obj_filter) == 0:
                    if invert:
                        single_subject_filter_inverted(kr, kw, rw, subj_idx, subj_filter)
                    else:
                        single_subject_filter(kr, kw, rw, subj_idx, subj_filter)

                elif len(subj_filter) == 0 and len(pred_filter) == 1 and len(obj_filter) == 0:
                    if invert:
                        single_predicate_filter_inverted(kr, kw, rw, pred_idx, pred_filter)
                    else:
                        single_predicate_filter(kr, kw, rw, pred_idx, pred_filter)

                elif len(subj_filter) == 0 and len(pred_filter) == 0 and len(obj_filter) == 1:
                    if invert:
                        single_object_filter_inverted(kr, kw, rw, obj_idx, obj_filter)
                    else:
                        single_object_filter(kr, kw, rw, obj_idx, obj_filter)
                else:
                    single_general_filter(kr, kw, rw, subj_idx, subj_filter, pred_idx, pred_filter, obj_idx, obj_filter)

            else:
                n_subj_filters: int = 0
                n_pred_filters: int = 0
                n_obj_filters: int = 0
                fidx: int
                for fidx in range(nfilters):
                    n_subj_filters += len(subj_filters[fidx])
                    n_pred_filters += len(pred_filters[fidx])
                    n_obj_filters += len(obj_filters[fidx])

                if n_subj_filters> 0 and n_pred_filters == 0 and n_obj_filters == 0 and first_match_only and not invert:
                    dispatch_subject_filter(kr, kws, rw, subj_idx, subj_filters)

                elif n_subj_filters == 0 and n_pred_filters > 0 and n_obj_filters == 0 and first_match_only and not invert:
                    dispatch_predicate_filter(kr, kws, rw, pred_idx, pred_filters)

                elif n_subj_filters == 0 and n_pred_filters == 0 and n_obj_filters > 0 and first_match_only and not invert:
                    dispatch_object_filter(kr, kws, rw, obj_idx, obj_filters)
                
                else:
                    multiple_general_filter(kr, kws, rw, subj_idx, subj_filters, pred_idx, pred_filters, obj_idx, obj_filters)

        finally:
            if verbose:
                print("Closing output files.", file=error_file, flush=True)
            for kw in kws:
                kw.close()
            if rw is not None:
                rw.close()
            if verbose:
                print("All output files have been closed.", file=error_file, flush=True)

        return 0

    def multiple_general_regex_fullmatch(kr: KgtkReader,
                                         kws: typing.List[KgtkWriter],
                                         rw: typing.Optional[KgtkWriter],
                                         subj_idx: int,
                                         subj_filters: typing.List[typing.Optional[typing.Pattern]],
                                         pred_idx: int,
                                         pred_filters: typing.List[typing.Optional[typing.Pattern]],
                                         obj_idx: int,
                                         obj_filters: typing.List[typing.Optional[typing.Pattern]]):
        if verbose:
            print("Applying a multiple-output general regex fullmatch filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Optional[typing.Pattern] = subj_filters[idx]
                pred_filter: typing.Optional[typing.Pattern] = pred_filters[idx]
                obj_filter: typing.Optional[typing.Pattern] = obj_filters[idx]
                idx += 1
                

                keep: bool = False
                reject: bool = False 
                if subj_filter is not None:
                    if len(row) <= subj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, subj_idx=%d): %s" % (input_line_count, idx, len(row), subj_idx, repr(row)))
                    subj_match: typing.Optional[typing.Match] = subj_filter.fullmatch(row[subj_idx])
                    if subj_match is not None:
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if pred_filter is not None:
                    if len(row) <= pred_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, pred_idx=%d): %s" % (input_line_count, idx, len(row), pred_idx, repr(row)))
                    pred_match: typing.Optional[typing.Match] = pred_filter.fullmatch(row[pred_idx])
                    if pred_match is not None:
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if obj_filter is not None:
                    if len(row) <= obj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, obj_idx=%d): %s" % (input_line_count, idx, len(row), obj_idx, repr(row)))
                    obj_match: typing.Optional[typing.Match] = obj_filter.fullmatch(row[obj_idx])
                    if obj_match is not None:
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def multiple_general_regex_match(kr: KgtkReader,
                                     kws: typing.List[KgtkWriter],
                                     rw: typing.Optional[KgtkWriter],
                                     subj_idx: int,
                                     subj_filters: typing.List[typing.Optional[typing.Pattern]],
                                     pred_idx: int,
                                     pred_filters: typing.List[typing.Optional[typing.Pattern]],
                                     obj_idx: int,
                                     obj_filters: typing.List[typing.Optional[typing.Pattern]]):
        if verbose:
            print("Applying a multiple-output general regex match filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Optional[typing.Pattern] = subj_filters[idx]
                pred_filter: typing.Optional[typing.Pattern] = pred_filters[idx]
                obj_filter: typing.Optional[typing.Pattern] = obj_filters[idx]
                idx += 1
                

                keep: bool = False
                reject: bool = False 
                if subj_filter is not None:
                    if len(row) <= subj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, subj_idx=%d): %s" % (input_line_count, idx, len(row), subj_idx, repr(row)))
                    subj_match: typing.Optional[typing.Match] = subj_filter.match(row[subj_idx])
                    if subj_match is not None:
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if pred_filter is not None:
                    if len(row) <= pred_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, pred_idx=%d): %s" % (input_line_count, idx, len(row), pred_idx, repr(row)))
                    pred_match: typing.Optional[typing.Match] = pred_filter.match(row[pred_idx])
                    if pred_match is not None:
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if obj_filter is not None:
                    if len(row) <= obj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, obj_idx=%d): %s" % (input_line_count, idx, len(row), obj_idx, repr(row)))
                    obj_match: typing.Optional[typing.Match] = obj_filter.match(row[obj_idx])
                    if obj_match is not None:
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def multiple_general_regex_search(kr: KgtkReader,
                                      kws: typing.List[KgtkWriter],
                                      rw: typing.Optional[KgtkWriter],
                                      subj_idx: int,
                                      subj_filters: typing.List[typing.Optional[typing.Pattern]],
                                      pred_idx: int,
                                      pred_filters: typing.List[typing.Optional[typing.Pattern]],
                                      obj_idx: int,
                                      obj_filters: typing.List[typing.Optional[typing.Pattern]]):
        if verbose:
            print("Applying a multiple-output general regex search filter", file=error_file, flush=True)

        input_line_count: int = 0
        reject_line_count: int = 0
        output_line_count: int = 0
        subj_filter_keep_count: int = 0
        pred_filter_keep_count: int = 0
        obj_filter_keep_count: int = 0
        subj_filter_reject_count: int = 0
        pred_filter_reject_count: int = 0
        obj_filter_reject_count: int = 0

        row: typing.List[str]
        for row in kr:
            input_line_count += 1

            written: bool = False

            idx: int = 0
            for kw in kws:
                subj_filter: typing.Optional[typing.Pattern] = subj_filters[idx]
                pred_filter: typing.Optional[typing.Pattern] = pred_filters[idx]
                obj_filter: typing.Optional[typing.Pattern] = obj_filters[idx]
                idx += 1
                

                keep: bool = False
                reject: bool = False 
                if subj_filter is not None:
                    if len(row) <= subj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, subj_idx=%d): %s" % (input_line_count, idx, len(row), subj_idx, repr(row)))
                    subj_match: typing.Optional[typing.Match] = subj_filter.search(row[subj_idx])
                    if subj_match is not None:
                        keep = True
                        subj_filter_keep_count += 1
                    else:
                        reject = True
                        subj_filter_reject_count += 1

                if pred_filter is not None:
                    if len(row) <= pred_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, pred_idx=%d): %s" % (input_line_count, idx, len(row), pred_idx, repr(row)))
                    pred_match: typing.Optional[typing.Match] = pred_filter.search(row[pred_idx])
                    if pred_match is not None:
                        keep = True
                        pred_filter_keep_count += 1
                    else:
                        reject = True
                        pred_filter_reject_count += 1

                if obj_filter is not None:
                    if len(row) <= obj_idx:
                        raise ValueError("Line %d: filter %d: short(len(row)=%d, obj_idx=%d): %s" % (input_line_count, idx, len(row), obj_idx, repr(row)))
                    obj_match: typing.Optional[typing.Match] = obj_filter.search(row[obj_idx])
                    if obj_match is not None:
                        keep = True
                        obj_filter_keep_count += 1
                    else:
                        reject = True
                        obj_filter_reject_count += 1

                if (keep if or_pattern else not reject) ^ invert:
                    kw.write(row)
                    if not written:
                        output_line_count += 1 # Count this only once.
                        written = True
                        if first_match_only:
                            break
                    
            if not written:
                if rw is not None:
                    rw.write(row)
                reject_line_count += 1

        if verbose:
            print("Read %d rows, rejected %d rows, wrote %d rows." % (input_line_count, reject_line_count, output_line_count), file=error_file, flush=True)
            print("Keep counts: subject=%d, predicate=%d, object=%d." % (subj_filter_keep_count, pred_filter_keep_count, obj_filter_keep_count), file=error_file, flush=True)
            print("Reject counts: subject=%d, predicate=%d, object=%d." % (subj_filter_reject_count, pred_filter_reject_count, obj_filter_reject_count), file=error_file, flush=True)

    def process_regex()->int:
        subj_regexes: typing.List[typing.Optional[typing.Pattern]] = [ ]
        pred_regexes: typing.List[typing.Optional[typing.Pattern]] = [ ]
        obj_regexes: typing.List[typing.Optional[typing.Pattern]] = [ ]

        subj_regex: typing.Optional[typing.Pattern]
        pred_regex: typing.Optional[typing.Pattern]
        obj_regex: typing.Optional[typing.Pattern]

        subj_needed: bool = False
        pred_needed: bool = False
        obj_needed: bool = False

        nfilters: int = 0
        pattern_list: typing.List[str]
        pattern: str
        for pattern_list in patterns:
            for pattern in pattern_list:
                subpatterns: typing.List[str] = pattern.split(pattern_separator)
                if len(subpatterns) != 3:
                    print("Error: The pattern must have three sections separated by %s (two %s total)." % (repr(pattern_separator), repr(pattern_separator)),
                          file=error_file, flush=True)
                    raise KGTKException("Bad pattern")
            
                subj_regex = prepare_regex(subpatterns[0])
                pred_regex = prepare_regex(subpatterns[1])
                obj_regex = prepare_regex(subpatterns[2])

                if subj_regex is None and pred_regex is None and obj_regex is None:
                    if verbose:
                        print("Warning: the filter %s is empty." % repr(pattern), file=error_file, flush=True)
                else:
                    subj_regexes.append(subj_regex)
                    pred_regexes.append(pred_regex)
                    obj_regexes.append(obj_regex)
                    nfilters += 1

                    if subj_regex is not None:
                        subj_needed = True
                    if pred_regex is not None:
                        pred_needed = True
                    if obj_regex is not None:
                        obj_needed = True

        if nfilters == 0:
            raise KGTKException("No filters found.")

        if nfilters != len(output_kgtk_files):
            if verbose:
                print("output files: %s" % " ".join([str(x) for x in output_kgtk_files]), file=error_file, flush=True)
            raise KGTKException("There were %d filters and %d output files." % (nfilters, len(output_kgtk_files)))

        if verbose:
            print("Opening the input file: %s" % str(input_kgtk_file), file=error_file, flush=True)
        kr: KgtkReader = KgtkReader.open(input_kgtk_file,
                                         options=reader_options,
                                         value_options = value_options,
                                         error_file=error_file,
                                         verbose=verbose,
                                         very_verbose=very_verbose,
        )

        subj_idx: int = kr.get_node1_column_index(subj_col)
        pred_idx: int = kr.get_label_column_index(pred_col)
        obj_idx: int = kr.get_node2_column_index(obj_col)

        # Complain about a missing column only when it is needed by the pattern.
        trouble: bool = False
        if subj_idx < 0 and subj_needed:
            trouble = True
            print("Error: Cannot find the subject column '%s'." % kr.get_node1_canonical_name(subj_col), file=error_file, flush=True)
        if pred_idx < 0 and pred_needed:
            trouble = True
            print("Error: Cannot find the predicate column '%s'." % kr.get_label_canonical_name(pred_col), file=error_file, flush=True)
        if obj_idx < 0 and obj_needed:
            trouble = True
            print("Error: Cannot find the object column '%s'." % kr.get_node2_canonical_name(obj_col), file=error_file, flush=True)
        if trouble:
            # Clean up:
            kr.close()
            raise KGTKException("Missing columns.")

        kw: KgtkWriter
        kws: typing.List[KgtkWriter] = [ ]
        output_kgtk_file: Path
        for output_kgtk_file in output_kgtk_files:
            if verbose:
                print("Opening the output file: %s" % str(output_kgtk_file), file=error_file, flush=True)
            kw = KgtkWriter.open(kr.column_names,
                                 output_kgtk_file,
                                 mode=KgtkWriter.Mode[kr.mode.name],
                                 use_mgzip=reader_options.use_mgzip, # Hack!
                                 mgzip_threads=reader_options.mgzip_threads, # Hack!
                                 error_file=error_file,
                                 verbose=verbose,
                                 very_verbose=very_verbose)
            kws.append(kw)

        rw: typing.Optional[KgtkWriter] = None
        if reject_kgtk_file is not None:
            if verbose:
                print("Opening the reject file: %s" % str(reject_kgtk_file), file=error_file, flush=True)
            rw = KgtkWriter.open(kr.column_names,
                                 reject_kgtk_file,
                                 mode=KgtkWriter.Mode[kr.mode.name],
                                 use_mgzip=reader_options.use_mgzip, # Hack!
                                 mgzip_threads=reader_options.mgzip_threads, # Hack!
                                 error_file=error_file,
                                 verbose=verbose,
                                 very_verbose=very_verbose)

        try:
            if match_type == "fullmatch":
                multiple_general_regex_fullmatch(kr, kws, rw, subj_idx, subj_regexes, pred_idx, pred_regexes, obj_idx, obj_regexes)
            elif match_type == "match":
                multiple_general_regex_match(kr, kws, rw, subj_idx, subj_regexes, pred_idx, pred_regexes, obj_idx, obj_regexes)
            elif match_type == "search":
                multiple_general_regex_search(kr, kws, rw, subj_idx, subj_regexes, pred_idx, pred_regexes, obj_idx, obj_regexes)
            else:
                raise KGTKException("Unknown match type %s" % repr(match_type))

        finally:
            if verbose:
                print("Closing output files.", file=error_file, flush=True)
            for kw in kws:
                kw.close()
            if rw is not None:
                rw.close()
            if verbose:
                print("All output files have been closed.", file=error_file, flush=True)

        return 0

    try:
        if regex:
            return process_regex()
        else:
            return process_plain()

    except Exception as e:
        kgtk_exception_auto_handler(e)
        return 1



import sys
from functools import lru_cache
import sqlite3
import threading
import time
import re
import io
import csv
import shlex

import kgtk.kypher.query as kyquery
import kgtk.cli.query as cliquery
import kgtk.kypher.sqlstore as sqlstore
from   kgtk.cli_argparse import KGTKArgumentParser
from   kgtk.exceptions import KGTKException

# avoid KGTK dependency on pandas and only import them if needed:
_have_pandas = False
def _import_pandas():
    global _have_pandas
    mod = sys.modules[__name__]
    import pandas as pd
    setattr(mod, "pd", pd)
    _have_pandas = True


# NOTES:
# - sqlite3 module already caches parsed SQL queries (100 by default),
#   so there is no need for prepared statements
# - executemany cannot be used to handle sets of input parameters,
#   it only works for modification commands, not selects


DEFAULT_GRAPH_CACHE_FILE = cliquery.DEFAULT_GRAPH_CACHE_FILE
DEFAULT_INDEX_MODE = 'auto'
DEFAULT_LOG_LEVEL = 0
DEFAULT_MAX_RESULTS = 100000
DEFAULT_MAX_CACHE_SIZE = 1000

DEFAULT_CONFIG = {
    # defaults for API configuration parameters:
    'GRAPH_CACHE'        : DEFAULT_GRAPH_CACHE_FILE,
    'INDEX_MODE'         : DEFAULT_INDEX_MODE,
    'MAX_RESULTS'        : DEFAULT_MAX_RESULTS,
    'MAX_CACHE_SIZE'     : DEFAULT_MAX_CACHE_SIZE,
    'LOG_LEVEL'          : DEFAULT_LOG_LEVEL,
}


# Command syntax parsing support

def define_query_command_parser(parser=None):
    """Define an argument parser that can parse KGTK query command-style definitions.
    """
    parser = parser or KGTKArgumentParser()
    # duplicate some of the logic in cli_entry which doesn't have a command parser factory:
    parser.add_argument('--debug', dest='_debug', action='store_true', help='enable debug mode')
    parser.add_argument('--expert', dest='_expert', action='store_true', help='enable expert mode')
    # add standard query argument definitions:
    cliquery.add_arguments_extended(parser, None)
    # define some additional API-only command options:
    parser.add_argument('--name', default=None, action='store', dest='name',
                        help="name to use for this query")
    parser.add_argument('--documentation', '--doc', default=None, action='store', dest='docstring',
                        help="documentation string for this query")
    parser.add_argument('--max-cache', '--maxcache', default=None, action='store', type=int, dest='maxcache',
                        help="maximal number of results to cache, 0 disables caching")
    parser.add_argument('--log-level', '--loglevel', default=None, action='store', type=int, dest='loglevel',
                        help="log level to use for this query")
    return parser

COMMAND_ARGUMENT_PARSER = define_query_command_parser()

def parse_query_command(command):
    """Parse a query 'command' into a normalized options dictionary.
    'command' may be a single string or an iterable of parsed arguments
    and may optionally contain 'kgtk' and 'query' particles if one
    wants to mirror top-level KGTK query commands exactly.
    """
    args = None
    if isinstance(command, str):
        args = shlex.split(command)
    elif command is not None:
        args = [str(arg) for arg in command]
    if not args:
        return {}
    
    parsed_args, rest_args = COMMAND_ARGUMENT_PARSER.parse_known_args(args)
    for rarg in rest_args:
        if rarg not in ('kgtk', 'query'):
            raise KGTKException(f'illegal query API option: {rarg}')
    # convert to dict:
    parsed_args = vars(parsed_args)
    if parsed_args.get('output') not in (None, '-'):
        raise KGTKException('output specification not supported in query API')
    loglevel = parsed_args.get('loglevel')
    options = cliquery.preprocess_query_options(**parsed_args)
    # if loglevel was set explicitly, override with it:
    if loglevel is not None:
        options['loglevel'] = loglevel
    elif options.get('loglevel') == 0:
        # undo the default in favor of API default:
        options['loglevel'] = None
    # remove all None values so we can use 'get' with defaults:
    return {k: v for k, v in options.items() if v is not None}


class KypherQuery(object):
    """
    Cachable Kypher query object.  Caches query translations and
    (if requested) query results.  Methods will generally not be called
    directly but through the linked API object.
    """

    def __init__(self, api, **kwargs):
        """Create a query object linked to the KypherApi 'api'.  All other
        arguments are passed to 'KypherQuery._define' (which see).
        """
        if not isinstance(api, KypherApi):
            raise KGTKException('query needs to be linked to existing API object')
        self.api = api
        self.kgtk_query = None
        self.sql = None
        self.parameters = None
        self.exec_wrapper = None
        self.definition_args = kwargs
        self.timestamp = -1
        self._define(**kwargs)

    def _define(self,
                inputs=None, doc=None,
                name=None, maxcache=None,
                query=None,
                match='()', where=None,
                opt=None, owhere=None,
                opt2=None, owhere2=None,
                with_='*', wwhere=None,
                ret='*', order=None,
                skip=None, limit=None,
                multi=None,
                parameters={},
                force=False,
                index=None, loglevel=None,
                cmd=None,
                **kwargs):
        """Internal constructor which generates a cached query translation and a
        LRU-cachable results structure (if requested).  See 'KypherApi.get_query'
        for documentation of arguments.
        """
        
        if self.kgtk_query is not None:
            raise KGTKException('query has already been defined')
        
        # any options specified in 'cmd' dominate others or get merged with them:
        cmd = parse_query_command(cmd)
        if name is not None and cmd.get('name', name) != name:
            raise KGTKException('multiply defined mismatched query names')
        name = cmd.get('name', name)

        inputs = cmd.get('input_files') or kyquery.listify(inputs) or self.api.get_all_inputs()
        norm_inputs = []
        for inp in inputs:
            if self.api.get_input_info(inp) is None:
                self.api.add_input(inp)
            norm_inputs.append(self.api.get_input(inp))
        inputs = norm_inputs

        self.docstring = cmd.get('docstring') or doc
        
        optionals = [(self._subst_graph_handles(opt, inputs), where) for opt, where in cmd.get('optionals', [])]
        opt and optionals.append((self._subst_graph_handles(opt, inputs), owhere))
        opt2 and optionals.append((self._subst_graph_handles(opt2, inputs), owhere2))
        # kwargs is an ordered dict, so the actual suffixes do not matter:
        for key, value in kwargs.items():
            if key.startswith('opt'):
                optionals.append([self._subst_graph_handles(value, inputs), None])
            elif key.startswith('owhere'):
                optionals[-1][1] = value
            else:
                raise KGTKException('Unexpected keyword argument: %s' % key)

        store = self.api.get_sql_store()
        loglevel = cmd.get('loglevel', loglevel if loglevel is not None else self.api.loglevel)
        index = cmd.get('index_mode', index or self.api.index_mode)
        # since we are loading results into memory, we are not using unlimited as the default:
        limit = cmd.get('limit', limit if limit is not None else self.api.max_results)
        # -1 forces unlimited results:
        if limit in (-1, '-1'):
            limit = None
        maxcache = cmd.get('maxcache', maxcache if maxcache is not None else self.api.max_cache_size)
        
        self.kgtk_query = kyquery.KgtkQuery (
            inputs, store,
            loglevel=loglevel, index=index,
            query=cmd.get('query', query),
            match=self._subst_graph_handles(cmd.get('match', match), inputs),
            where=cmd.get('where', where),
            optionals=optionals,
            with_=cmd.get('with', (with_, wwhere)),
            ret=cmd.get('return', ret),
            order=cmd.get('order', order),
            skip=cmd.get('skip', skip),
            limit=limit,
            multi=cmd.get('multi_edge', multi),
            parameters=cmd.get('parameters', parameters),
            force=cmd.get('force', force))
        
        self.kgtk_query.defer_params = True
        state = self.kgtk_query.translate_to_sql()
        self.sql = state.get_sql()
        self.parameters = state.get_parameters()
        self.kgtk_query.ensure_relevant_indexes(state)
        # create memoizable execution wrapper:
        self.exec_wrapper = lambda q, p, f: q._exec(p, f)
        if maxcache > 0:
            self.exec_wrapper = lru_cache(maxsize=maxcache)(self.exec_wrapper)
        if name is not None:
            self.api.cached_queries[name] = self
        self.timestamp = self.api.timestamp
        return self

    PSEUDO_GRAPH_HANDLE_REGEX = re.compile('`?\$(?P<handle>[^$`:\s]+)`?:')

    def _subst_graph_handles(self, match_pattern, inputs):
        """Substitute any pseudo graph handles in 'match_pattern' that are relevant to one
        of 'inputs' with their canonical input file or alias (backtick quoted if necessary).
        Any backtick quotes around a handle will be dropped automatically first.
        """
        cursor = 0
        out = io.StringIO()
        for match in self.PSEUDO_GRAPH_HANDLE_REGEX.finditer(match_pattern):
            handle = match.group('handle')
            info = self.api.get_input_info(handle)
            if info is not None and info['handle']:
                # we found a match for a pseudo-handle, check whether it is relevant:
                for inp in inputs:
                    if self.api.get_input_info(inp) == info:
                        # handle is for one of the listed 'inputs':
                        out.write(match_pattern[cursor:match.start()])
                        cursor = match.end()
                        out.write('`%s`:' % self.api.get_input(handle))
                        break
        if cursor == 0:
            return match_pattern
        else:
            out.write(match_pattern[cursor:])
            return out.getvalue()

    def clear(self):
        """Clear all currently cached information for this query.
        Subsequent executions will regenerate everything as needed.
        """
        self.kgtk_query = None
        self.sql = None
        self.parameters = None
        self.exec_wrapper = None
        self.timestamp = -1

    def refresh(self):
        """Check whether cached values are still valid relative to the current state of the API.
        If not, regenerate all cached structures according to the stored definition arguments.
        """
        if self.timestamp < self.api.timestamp:
            self.clear()
            self._define(**self.definition_args)

    def _subst_params(self, params, substitutions):
        """Return a copy of the list 'params' modified by any 'substitutions'.
        Placeholder parameters in 'params' are marked as single-element tuples, e.g., ('NODE').
        """
        subst_params = []
        # we can't do this easily with a list comprehension, since False or 0 are valid substitutions:
        for x in params:
            if isinstance(x, tuple):
                subst_params.append(substitutions.get(x[0], x[0]))
            else:
                subst_params.append(x)
        # we return the result as a tuple which can be hashed for memoization:
        return tuple(subst_params)

    def _exec(self, parameters, fmt):
        """Internal query execution wrapper that can easily be memoized.
        """
        # TO DO: abstract some of this better in KgtkQuery API
        kgtk_query = self.kgtk_query
        result = kgtk_query.store.execute(self.sql, parameters)
        if kgtk_query.result_header is None or kgtk_query.multi_edge is not None:
            # NOTE: if we have multi edges, we need to recompute the original full-width header:
            kgtk_query.result_header = [kgtk_query.unalias_column_name(c[0]) for c in result.description]
            result = kgtk_query.wrap_multi_edge_result(result)
        if fmt is None:
            # convert to list so we can reuse if we memoize:
            return tuple(result)
        # allow types and their names:
        fmt = hasattr(fmt, '__name__') and fmt.__name__ or str(fmt)
        if fmt == 'iter':
            return result
        elif fmt == 'tuple':
            return tuple(result)
        elif fmt == 'list':
            return list(result)
        elif fmt in ('df', 'dataframe', 'DataFrame'):
            if not _have_pandas:
                _import_pandas()
            return pd.DataFrame(result, columns=kgtk_query.result_header)
        # TO DO: consider supporting namedtuple and/or sqlite3.Row as row_factory types
        #        (for sqlite3.Row we have the issue that aliases become keys())
        else:
            raise KGTKException('unsupported query result format: %s' % fmt)

    def execute(self, fmt=None, **params):
        """Execute this query with the given 'params' and return the result in format 'fmt'. 
        By default the result is a list of tuples (see 'KypherApi.execute_query' for other options).
        'params' should be a list of key/value pairs for the unbound Kypher parameters in the query.
        For example, the Kypher parameter '$NODE' can be bound with 'NODE=<some value>'.
        """
        self.refresh()
        parameters = self._subst_params(self.parameters, params)
        result = self.exec_wrapper(self, parameters, fmt)
        return result

    def execute_to_file(self, file=sys.stdout, noheader=False, **params):
        """Execute this query with the given 'params' and write the result to the file or
        file-like object 'file' in KGTK format.  Output a header unless 'noheader' is true.
        """
        self.refresh()
        if hasattr(self.exec_wrapper, 'cache_clear'):
            # if this is a re-call of a caching query, ensure it is cleared,
            # since the result iterator was used up in the previous call:
            self.exec_wrapper.cache_clear()
        parameters = self._subst_params(self.parameters, params)
        try:
            out = open(file, 'w') if isinstance(file, str) else file
            if not hasattr(out, 'write'):
                raise KGTKException('expected file or file-like object')
            result = self.exec_wrapper(self, parameters, iter)
            csvwriter = csv.writer(out, dialect=None, delimiter='\t',
                                   quoting=csv.QUOTE_NONE, quotechar=None,
                                   lineterminator='\n', escapechar=None)
            if not noheader:
                csvwriter.writerow(self.get_result_header())
            csvwriter.writerows(result)
        finally:
            if isinstance(file, str):
                out.close()

    def get_result_header(self, error=True):
        """Return the list of column names for this query.  This requires the query to have
        run at least once (also again after caches were cleared).
        """
        self.refresh()
        header = self.kgtk_query.result_header
        if header is None and error:
            raise KGTKException('query needs to be run at least once to access its result header')
        return header


class KypherApi(object):
    """
    Kypher query API.  Manages a connection to a single graph cache DB,
    caches translated queries and (if requested) query results.
    The API can be configured with the constructor or by passing in a
    configuration object which can also be used to store additional
    user-defined configuration keys and values.

    Many of the API options are direct analogs of corresponding options of the
    KGTK 'query' command.  Please consult the 'query' manual for more details.
    """

    def __init__(self,
                 graphcache=None,
                 index=None,
                 maxresults=None,
                 maxcache=None,
                 loglevel=None,
                 config=None,
                 readonly=False):
        """Create a new API object and initialize a number of configuration values.
        'graphcache' should be a filename for a Kypher graph cache to create or reuse.
        It uses the same default as the --graph-cache option of the 'query' command.

        'index' specifies the default indexing mode which defaults to the value of the
        'INDEX_MODE' property in the provided user or default configuration.  It uses
        the same default value as the --index option of the 'query' command.  'index'
        can be overridden for individual queries.

        'maxresults' specifies the default maximum number of results to be returned by
        a query which defaults to the 'MAX_RESULTS' property in the provided user or default 
        configuration.  Different from the --limit option of the 'query' command, API query
        results are not unbounded by default (since they might exhaust available RAM).  Use
        -1 to allow unbounded results.  'maxresults' can be overridden for individual queries.

        'maxcache' specifies the default number of per-query results to memoize in an LRU
        cache which defaults to the 'MAX_CACHE_SIZE' property in the provided user or default 
        configuration.  Setting this to 0 disables query result caching.  'maxcache' can be
        overridden for individual queries.

        'loglevel' controls the verbosity of logging output.  It defaults to the value of the
        'LOG_LEVEL' property in the provided user or default configuration.  The default is 0
        and values 1 and 2 correspond to --debug and --expert modes of the 'query' command.
        'loglevel' can be overridden for individual queries.

        'config' can be used to initialize configuration values from a user-provided
        configuration object which may have additional user-defined key/value pairs.
        Standard configuration values for the keys described above will default to their
        values in 'DEFAULT_CONFIG' if no user-defined values are provided.

        'readonly' controls whether the graph cache is opened in read-only mode or not.
        For read-only, the graph cache must exist and contain data that is properly indexed.
        """
        self.config = config or {}
        self.graph_cache = graphcache
        if graphcache is None:
            self.graph_cache = self.get_config('GRAPH_CACHE')
        self.readonly = readonly
        self.index_mode = index
        if index is None:
            self.index_mode = self.get_config('INDEX_MODE')
        self.max_results = maxresults
        if maxresults is None:
            self.max_results = self.get_config('MAX_RESULTS')
        self.max_cache_size = maxcache
        if maxcache is None:
            self.max_cache_size = self.get_config('MAX_CACHE_SIZE')
        self.sql_store = None
        self.lock = threading.Lock()
        self.loglevel = loglevel
        if loglevel is None:
            self.loglevel = self.get_config('LOG_LEVEL')
        self.timestamp = -1
        self._initialize()

    def _initialize(self):
        """Re/initialize internal data structures.
        """
        if self.sql_store is not None:
            self.close()
        self.inputs = {}
        self.cached_queries = {}
        self.timestamp += 1

    def close(self):
        """Close the connection to the database and clear all caches and information.
        """
        if self.sql_store is not None:
            self.sql_store.close()
            # avoid recursion in '_initialize':
            self.sql_store = None
        self.clear_caches()
        self._initialize()

    def clear(self):
        """Close the connection to the database and clear all caches and information.
        """
        self.close()

    def clear_caches(self):
        """Clear all query translation and LRU result caches.  Queries that were not given
        a name but were assigned to any variables or slots will rebuild automatically upon next
        execution.  However, to release their storage they will need to be cleared manually
        with 'q.clear()' or disconnected from variables or objects pointing to them.
        """
        klass = type(self)
        for name in dir(klass):
            obj = getattr(klass, name, None)
            if hasattr(obj, 'cache_clear'):
                obj.cache_clear()
        for query in self.cached_queries.values():
            query.clear()
        self.timestamp += 1

    def clear_queries(self):
        """Clear all currently defined queries and any caches depending on them.
        """
        self.clear_caches()
        self.cached_queries = {}
        
    def clear_inputs(self):
        """Clear all currently defined inputs and any caches depending on them.
        """
        self.clear_caches()
        self.inputs = {}


    def get_config(self, key, dflt=None):
        """Access a configuration value for 'key' with default 'dflt'.
        """
        dflt = dflt or DEFAULT_CONFIG.get(key)
        return self.config.get(key, dflt)

    def set_config(self, key, value):
        """Set the configuration value for 'key' to 'value'.
        This will also reset all query and LRU caches, since they might be specific
        to a particular configuration value.
        """
        self.clear_caches()
        self.config[key] = value

    def log(self, level, message):
        """Output log 'message' if the current log level is at least 'level'.
        """
        if self.loglevel >= level:
            header = '[%s query]:' % time.strftime('%Y-%m-%d %H:%M:%S')
            sys.stderr.write('%s %s\n' % (header, message))
            sys.stderr.flush()
    
    def get_sql_store(self):
        """Create a new SQL store object from the configured graph_cache file or return a cached value.
        """
        if self.sql_store is None:
            if self.readonly:
                conn = sqlite3.connect(f'file:{self.graph_cache}?mode=ro', uri=True, check_same_thread=False)
            else:
                conn = sqlite3.connect(self.graph_cache, check_same_thread=False)
            # don't do this for now, since we get the aliases as keys() which would require further mapping:
            #conn.row_factory = sqlite3.Row
            self.sql_store = sqlstore.SqliteStore(dbfile=self.graph_cache, conn=conn,
                                                  loglevel=self.loglevel, readonly=self.readonly)
        return self.sql_store

    def get_lock(self):
        """Return the lock object.
        """
        return self.lock

    def __enter__(self):
        """Lock context manager for 'with ... as api:' idiom.
        """
        if self.get_lock().locked():
            # for now for debugging:
            print('Waiting for Kypher API lock...')
        self.get_lock().acquire()
        return self

    def __exit__(self, *_exc):
        self.get_lock().release()

    def get_input_info(self, name):
        """Return the info structure associated with input 'name', which maybe an input file name,
        input file alias or API-local name.
        """
        return self.inputs.get(name)

    def add_input(self, file, alias=None, name=None, handle=False, load=False):
        """Add input 'file' as one of the input files known by this API instance.

        If 'alias' is not None, use it as the name of input inside the graph cache
        (equivalent to the --as option of the 'query' command).  

        If 'name' is given, use it as an additional API-local name to refer to the input.

        If 'handle' is True, allow 'alias' or 'name' to be used as a pseudo-graph handle
        in match patterns using the Kypher parameter syntax.  These handles are 'pseudo'
        only, since they get replaced with their full respective input file name *before*
        any Kypher parsing is done.  For example, if we have the name 'mygraph' defined
        for an input, it can be used as '$mygraph: (x)-[]->(y), ...' in a strict or
        optional match pattern.  The syntax of handle names is less restrictive than for
        real Kypher variables and can contain any character except whitespace and '$`:'.
        
        If 'load' is true, load the input file immediately, unless it is already in the
        graph cache.  Using this function is only needed if an input alias and/or API name
        is desired, or if data loading should be forced.  Otherwise, inputs can be specified 
        directly to 'get_query' (which see).  NOTE: providing an alias implies 'load=True'.
        """
        info = self.get_input_info(file) or self.get_input_info(alias) or self.get_input_info(name)
        info = info or {}
        info['file'] = file
        self.inputs[file] = info
        if alias is not None:
            info['alias'] = alias
            self.inputs[alias] = info
        if name is not None:
            info['name'] = name
            self.inputs[name] = info
        info['handle'] = handle
        if load or info.get('alias') is not None:
            # we have to preload if we use a DB alias, otherwise a rerun of a cached query will fail:
            store = self.get_sql_store()
            store.add_graph(file, alias=info.get('alias'))
        # clear query caches to avoid any file/alias confusions in cached queries:
        self.clear_caches()

    def get_input(self, key):
        """Get the canonical input named by 'key' so it can be used with 'get_query'.
        'key' maybe an input file name, input file alias or API-local name.
        """
        info = self.get_input_info(key)
        if info is None:
            raise KGTKException('no input named by key: %s' % key)
        inp = info.get('alias')
        if inp is None:
            inp = info.get('file')
        return inp

    def get_all_inputs(self):
        """Get the canonical names of all currently defined inputs.
        """
        inputs = set()
        for info in self.inputs.values():
            inputs.add(self.get_input(info['file']))
        return list(inputs)

    def lookup_query(self, name):
        """Return a cached query with 'name' if it exists, otherwise return None.
        This is useful as a cheap check before a fully parameterized call to 'get_query'.
        """
        return self.cached_queries.get(name)

    def _get_query(self, query, error=True):
        """Internal accessor that allows transparent 'query' access via objects or names.
        """
        if isinstance(query, KypherQuery):
            return query
        kypher_query = self.cached_queries.get(query)
        if kypher_query is not None:
            return kypher_query
        elif error:
            raise KGTKException('cannot find query with name: %s' % query)
        else:
            return None

    def get_query(self,
                  inputs=None, doc=None,
                  name=None, maxcache=None,
                  query=None,
                  match='()', where=None,
                  opt=None, owhere=None,
                  opt2=None, owhere2=None,
                  with_='*', wwhere=None,
                  ret='*', order=None,
                  skip=None, limit=None,
                  multi=None,
                  parameters={},
                  force=False,
                  index=None,
                  loglevel=None,
                  cmd=None,
                  **kwargs):
        
        """Construct or reuse a cached query.  Many of the options here directly correspond
        to one of the options of the KGTK 'query' command, so we only describe core differences.

        'inputs' is a single or list of input file names or aliases just as they would be supplied
        to the 'query' command.  If inputs were defined with aliases or API-internal names, 'get_input'
        should be used to access their canonical name to ensure the proper file name or alias usage.
        All inputs will automatically be added via 'add-input' (which is a no-op if they already exist).
        If 'inputs' is None, the list of all currently defined inputs will be supplied in random order.
        In that case, all match clauses need to explicitly specify the graph they apply to.

        'doc' can be used to attach a documentation string to the query.

        If 'name' is given, the constructed query object will be cached in the API object under that name.
        Methods such as 'execute_query' can then be called with a query name or object.  Even if no name is
        used, the query can still be executed multiple times by saving it to a variable or slot.
        If a cached query with 'name' already exists, 'get_query' returns that query immediately without
        any further translation and processing.  That means a named 'get_query' call can be used in a
        processing loop without incurring any notable overhead after the first call (besides assembling
        and passing the parameters to 'get_query').

        'maxcache' specifies the number of per-query results to memoize in an LRU cache which defaults
        to the number defined during creation of the API object.  To disable LRU caching for this query,
        set 'maxcache' to 0.  NOTE: the number specifies the maximum number of query results to cache,
        but each result may have an arbitrarily large number of rows, so the storage consumed by this
        cache might become potentially much larger than indicated by 'maxcache' alone.

        'query' can be a fully specified Kypher query (just as the --query option of the 'query' command).

        'match' is the strict match clause of a Kypher query (just as the --match option of the 'query'
        command).  Match clauses should be properly qualified with graph handles to refer to the appropriate
        input in 'inputs'.  Use 'get_input' to access the proper canonical input names.  For example:
        ... 'match="`%s`: (n1)-[r]->(n2)" % api.get_input("my_graph")', ...

        'where' is the where clause of a strict match clause (just as --where following --match for 'query').
        
        'opt' is the first optional match clause (just as --opt for 'query').  As with 'match', appropriate
        graph handles should generally be used.

        'owhere' is the where clause of the first optional match clause (just as --where following the first
        --opt clause in 'query').

        'opt2' is the second optional match clause (just as the second --opt for 'query').

        'owhere2' is the where clause of the second optional match clause (just as --where following the
        second --opt clause in 'query').

        Additional optional clauses can be specified with 'opt3', 'owhere3', 'opt4', 'owhere4', etc.  They
        will be ordered in their order of specification, not according to their suffixes which can really be
        arbitrary, as long as they have not been previously used.

        'with_' is the with clause of a Kypher query (just as the --with option of the 'query' command).

        'wwhere' is the where clause of the with clause (or global where) (just as the --where clause following
        --with or the --where: option of the 'query' command).

        'ret' is the return clause of a Kypher query (just as the --return option of the 'query' command).

        'order' is the order-by clause of a Kypher query (just as the --order-by option of the 'query' command).

        'skip' is the skip clause of a Kypher query (just as the --skip option of the 'query' command).

        'limit' is the limit clause of a Kypher query (just as the --limit option of the 'query' command).
        However, the default for 'limit' is not unbounded but the number of 'maxresults' configured during
        creation of the API object.  To specify unlimited results, use -1 as the value for 'limit'.

        'multi' is the multi-edge option of a Kypher query (just as the --multi option of the 'query' command).

        'parameters' is a fixed dictionary of Kypher parameter name/value pairs to use in each invocation
        of this query (parameter names do not include the leading dollar sign).  Any parameters whose values
        are not defined in this dictionary will need to be bound in each call to 'execute'.

        'force' specifies the value of the --force option in a 'query' command.

        'index' specifies the indexing mode to use for this query which defaults to the number defined during
        creation of the API object (just like the --index option of the 'query' command).

        'loglevel' controls the verbosity of logging output and defaults to the number defined during
        creation of the API object (similar to the --debug and --expert options of the 'query' command).

        'cmd' can be used to specify a query in KGTK query command syntax by giving either a complete string or
        a list of parsed options.  The accepted syntax is exactly the same as for the query shell command.
        Options specified here dominate other 'get_query' options or get merged with them where possible.
        """
        
        kypher_query = self._get_query(name, error=False)
        if kypher_query is not None:
            return kypher_query
        kypher_query = KypherQuery(
            self, inputs=inputs, doc=doc, name=name, maxcache=maxcache,
            query=query, match=match, where=where, opt=opt, owhere=owhere, opt2=opt2, owhere2=owhere2,
            with_=with_, wwhere=wwhere, ret=ret, order=order, skip=skip, limit=limit, multi=multi,
            parameters=parameters, force=force, index=index, loglevel=loglevel, cmd=cmd,
            **kwargs)
        return kypher_query

    def execute_query(self, query, fmt=None, **params):
        """Execute 'query' with the given 'params' and return the result in format 'fmt'
        ('query' may be a name or object).  By default the result is a tuple of tuples.
        Supported formats are 'iter' (the sqlite result iterator), 'tuple' (a tuple list
        of result tuples), 'list' (a list of tuples), or 'df' (a Pandas dataframe which 
        requires the pandas module to be available - 'dataframe' or 'DataFrame' are also
        supported as format names).  Type objects can also be used if available instead
        of their names.  IMPORTANT: If 'iter' is chosen, the result cannot effectively be
        cached, since the iterator will be exhausted after first traversal.
        'params' should be a list of key/value pairs for the unbound Kypher parameters in
        'query' (those not specified in the 'parameters' argument to 'get_query').
        For example, the Kypher parameter '$NODE' can be bound with 'NODE=<some value>'.
        """
        return self._get_query(query).execute(fmt=fmt, **params)

    def execute_query_to_file(self, query, file=sys.stdout, noheader=False, **params):
        """Execute 'query' with the given 'params' and write the result to the file or
        file-like object 'file' in KGTK format.  Output a header unless 'noheader' is true.
        """
        query.execute_to_file(file=file, noheader=noheader, **params)

    def get_query_result_header(self, query, error=True):
        """Return the list of column names for 'query' (which may be a name or object).
        This requires the query to have run at least once (also again after caches were cleared).
        An error is raised if this is called too early unless 'error' is false.
        """
        return self._get_query(query).get_result_header(error=error)


"""
# Example 1:

>>> import kgtk.kypher.api as kapi

# we don't supply a graph cache file, so the default in /tmp will be used:
>>> api = kapi.KypherApi()

# we define some inputs with API-local names for easy reference, but the
# filenames will be used as the actual input names inside the graph cache:
>>> api.add_input('examples/docs/query-graph.tsv', name='graph', handle=True)
>>> api.add_input('examples/docs/query-works.tsv', name='works', handle=True)
>>> api.add_input('examples/docs/query-quals.tsv', name='quals')

# This defines a query object which we will refer to by its name later.
# Since log level is 1, it displays the SQL translation it produces.
# This call only prepares the query, it does not actually run it.
# We use the Kypher parameter $ORG to later query for different orgs.
# Parameters cannot be used in the match clause, so we use it in 'where':
>>> api.get_query(name='work-info',
                  # we qualify each clause with a graph handle, since all inputs are supplied to the query by default;
                  # we use pseudo graph handles here which is easiest (using handle=True in add_input above):
                  match='$works: (p)-[r:works]->(c), $graph: (p)-[:name]->(n)',
                  where='c=$ORG',
                  # here we splice in the graph handle directly, just for illustration of the separate mechanism;
                  # note how we use backtick quoting to escape special characters in file names:
                  opt=  '`%s`: (r)-[:starts]->(s)' % api.get_input('quals'),
                  ret=  'c as company, p as employee, n as name, s as start',
                  loglevel=1)
... ... ... ... ... 
[2021-06-28 20:51:13 query]: SQL Translation:
---------------------------------------------
  SELECT graph_2_c1."node2" "_aLias.company", graph_2_c1."node1" "_aLias.employee", graph_3_c2."node2" "_aLias.name", graph_1_c3."node2" "_aLias.start"
     FROM graph_2 AS graph_2_c1
     INNER JOIN graph_3 AS graph_3_c2
     ON graph_2_c1."node1" = graph_3_c2."node1"
        AND graph_2_c1."label" = ?
        AND graph_3_c2."label" = ?
        AND (graph_2_c1."node2" = ?)
     LEFT JOIN graph_1 AS graph_1_c3
     ON graph_2_c1."id" = graph_1_c3."node1"
        AND graph_1_c3."label" = ?
     LIMIT ?
  PARAS: ['works', 'name', ('ORG',), 'starts', 100000]
---------------------------------------------
<kgtk.kypher.api.KypherQuery object at 0x7ff099bb0950>

# Now we run the query with a specific binding for $ORG.
# By default, the result is returned as a list of tuples:
>>> api.execute_query('work-info', ORG='ACME')
[('ACME', 'Hans', "'Hans'@de", '^1984-12-17T00:03:12Z/11')]

# Here we ask with a different binding and using a pandas
# dataframe as the result format (requires pandas to be installed):
>>> api.execute_query('work-info', fmt='df', ORG='Kaiser')
  company employee       name                     start
0  Kaiser     Otto  'Otto'@de                      None
1  Kaiser      Joe      "Joe"  ^1996-02-23T08:02:56Z/09

# Here we call 'get_query' with the same arguments as above.
# Since a cached query for 'work-info' already exists, it is used
# without requiring any additional translation (thus no log output).
# Since the query is asked with a binding we queried for before,
# the cached result set is used without any query execution at all.
# This time we use the 'execute' method on the query object directly:
>>> api.get_query(name='work-info',
                  match='$works: (p)-[r:works]->(c), $graph: (p)-[:name]->(n)',
                  where='c=$ORG',
                  opt=  '`%s`: (r)-[:starts]->(s)' % api.get_input('quals'),
                  ret=  'c as company, p as employee, n as name, s as start',
                  loglevel=1).execute(fmt='df', ORG='Kaiser')
... ... ... ... ...   
company employee       name                     start
0  Kaiser     Otto  'Otto'@de                      None
1  Kaiser      Joe      "Joe"  ^1996-02-23T08:02:56Z/09

# if we clear caches first, the query gets translated from scratch:
>>> api.clear_caches()
>>> api.get_query(name='work-info',
                  match='$works: (p)-[r:works]->(c), $graph: (p)-[:name]->(n)',
                  where='c=$ORG',
                  opt=  '`%s`: (r)-[:starts]->(s)' % api.get_input('quals'),
                  ret=  'c as company, p as employee, n as name, s as start',
                  loglevel=1).execute(fmt='df', ORG='Kaiser')
... ... ... ... ... 
[2021-06-29 12:40:01 query]: SQL Translation:
---------------------------------------------
  SELECT graph_2_c1."node2" "_aLias.company", graph_2_c1."node1" "_aLias.employee", graph_3_c2."node2" "_aLias.name", graph_1_c3."node2" "_aLias.start"
     FROM graph_2 AS graph_2_c1
     INNER JOIN graph_3 AS graph_3_c2
     ON graph_2_c1."node1" = graph_3_c2."node1"
        AND graph_2_c1."label" = ?
        AND graph_3_c2."label" = ?
        AND (graph_2_c1."node2" = ?)
     LEFT JOIN graph_1 AS graph_1_c3
     ON graph_2_c1."id" = graph_1_c3."node1"
        AND graph_1_c3."label" = ?
     LIMIT ?
  PARAS: ['works', 'name', ('ORG',), 'starts', 100000]
---------------------------------------------
  company employee       name                     start
0  Kaiser     Otto  'Otto'@de                      None
1  Kaiser      Joe      "Joe"  ^1996-02-23T08:02:56Z/09

# clearing caches preserves query definitions, so we don't have to define them again:
>>> api.clear_caches()
>>> api.execute_query('work-info', fmt='df', ORG='Kaiser')
[2021-06-29 13:10:03 query]: SQL Translation:
---------------------------------------------
  SELECT graph_2_c1."node2" "_aLias.company", graph_2_c1."node1" "_aLias.employee", graph_3_c2."node2" "_aLias.name", graph_1_c3."node2" "_aLias.start"
     FROM graph_2 AS graph_2_c1
     INNER JOIN graph_3 AS graph_3_c2
     ON graph_2_c1."node1" = graph_3_c2."node1"
        AND graph_2_c1."label" = ?
        AND graph_3_c2."label" = ?
        AND (graph_2_c1."node2" = ?)
     LEFT JOIN graph_1 AS graph_1_c3
     ON graph_2_c1."id" = graph_1_c3."node1"
        AND graph_1_c3."label" = ?
     LIMIT ?
  PARAS: ['works', 'name', ('ORG',), 'starts', 100000]
---------------------------------------------
  company employee       name                     start
0  Kaiser     Otto  'Otto'@de                      None
1  Kaiser      Joe      "Joe"  ^1996-02-23T08:02:56Z/09

>>> api.close()
>>>


# Example 2:

# here we supply the location of an existing Wikidata graph cache:
>>> api = kapi.KypherApi(graphcache='..../wikidata-20210215-dwd-browser.sqlite3.db', loglevel=1)

# we know it has a graph named 'claims', so we use that directly in the query instead of 'add_input()':
>>> query = api.get_query(inputs='claims',
                          name='get_node_edges',
                          maxcache=100,
                          match='(n)-[r]->(n2)',
                          where='n=$NODE',
                          ret='r as id, n as node1, r.label as label, n2 as node2')
... ... ... ... ... 
[2021-06-26 12:03:40 query]: SQL Translation:
---------------------------------------------
  SELECT graph_1_c1."id" "_aLias.id", graph_1_c1."node1" "_aLias.node1", graph_1_c1."label" "_aLias.label", graph_1_c1."node2" "_aLias.node2"
     FROM graph_1 AS graph_1_c1
     WHERE (graph_1_c1."node1" = ?)
     LIMIT ?
  PARAS: [('NODE',), 100000]
---------------------------------------------

# here we query with the query object instead of its name:
>>> api.execute_query(query, NODE='Q40', fmt='df')
                              id node1  label                                   node2
0    Q40-P1036-084620-8193d4f9-0   Q40  P1036                                "2--436"
1    Q40-P1081-0811b0-4f8b4eb7-0   Q40  P1081                                  +0.747
2    Q40-P1081-0937d9-77e2a41e-0   Q40  P1081                                  +0.861
3    Q40-P1081-132349-a529cefe-0   Q40  P1081                                  +0.835
4    Q40-P1081-1af44c-dacd9580-0   Q40  P1081                                  +0.764
..                           ...   ...    ...                                     ...
572   Q40-P949-f04fc2-1eaa977b-0   Q40   P949                             "000013315"
573   Q40-P950-3672f3-53465804-0   Q40   P950                              "XX450936"
574   Q40-P982-59891e-540db01d-0   Q40   P982  "caac77d1-a5c8-3e6e-8e27-90b44dcc1446"
575   Q40-P984-d87d6b-e2308567-0   Q40   P984                                   "AUT"
576   Q40-P998-05a2ec-b33435af-0   Q40   P998              "Regional/Europe/Austria/"

[577 rows x 4 columns]
>>> api.execute_query(query, NODE='Q41', fmt='df')
                              id node1  label                                   node2
0    Q41-P1036-682e2c-8ed5cb13-0   Q41  P1036                                "2--495"
1    Q41-P1081-02c2ed-926e1bf2-0   Q41  P1081                                  +0.866
2    Q41-P1081-0d7cc4-44b73d12-0   Q41  P1081                                  +0.762
3    Q41-P1081-11fb19-e2b6d2a2-0   Q41  P1081                                  +0.789
4    Q41-P1081-132349-6855d15e-0   Q41  P1081                                  +0.835
..                           ...   ...    ...                                     ...
594   Q41-P948-fc515d-8ee9b292-0   Q41   P948    "Kompsatos river bridge, Thrace.jpg"
595   Q41-P949-935996-53ecc964-0   Q41   P949                             "000979098"
596   Q41-P982-031a25-5b7e2e1a-0   Q41   P982  "803db0ca-b6ed-3bbc-aeb8-f89efd0a2168"
597   Q41-P984-f048da-9b23a812-0   Q41   P984                                   "GRE"
598   Q41-P998-b58b57-13f3de7d-0   Q41   P998               "Regional/Europe/Greece/"

[599 rows x 4 columns]
>>> api.execute_query(query, NODE='Q52353442')
[('Q52353442-P1006-235de5-5b24a9cf-0', 'Q52353442', 'P1006', '"321658140"'), ('Q52353442-P1015-cd5f62-c7e646d4-0', 'Q52353442', 'P1015', '"2100657"'), ('Q52353442-P106-Q82594-28c48a6c-0', 'Q52353442', 'P106', 'Q82594'), ('Q52353442-P108-Q37548-61855425-0', 'Q52353442', 'P108', 'Q37548'), ('Q52353442-P108-Q4614-dbb39b28-0', 'Q52353442', 'P108', 'Q4614'), ('Q52353442-P1153-bef441-003e2488-0', 'Q52353442', 'P1153', '"7004618158"'), ('Q52353442-P1416-Q6030821-708a6b3e-0', 'Q52353442', 'P1416', 'Q6030821'), ('Q52353442-P166-Q18748039-fc276466-0', 'Q52353442', 'P166', 'Q18748039'), ('Q52353442-P166-Q18748042-3c446357-0', 'Q52353442', 'P166', 'Q18748042'), ('Q52353442-P184-Q6123694-2032855e-0', 'Q52353442', 'P184', 'Q6123694'), ('Q52353442-P185-Q102343472-373f0ed5-0', 'Q52353442', 'P185', 'Q102343472'), ('Q52353442-P19-Q2807-92d8e80c-0', 'Q52353442', 'P19', 'Q2807'), ('Q52353442-P1960-601bc1-1d1f5261-0', 'Q52353442', 'P1960', '"NDmGsCgAAAAJ"'), ('Q52353442-P21-Q6581072-f653b32c-0', 'Q52353442', 'P21', 'Q6581072'), ('Q52353442-P213-49066c-8fc0d32c-0', 'Q52353442', 'P213', '"0000 0000 3762 2146"'), ('Q52353442-P214-1c576a-648c3580-0', 'Q52353442', 'P214', '"18334143"'), ('Q52353442-P244-5d352f-f85a8c96-0', 'Q52353442', 'P244', '"nb2005019080"'), ('Q52353442-P2456-b325a5-5b751548-0', 'Q52353442', 'P2456', '"88/2686"'), ('Q52353442-P2671-2e7fe8-0a9466a9-0', 'Q52353442', 'P2671', '"/g/11grpt57xn"'), ('Q52353442-P27-Q29-1c7b6b5c-0', 'Q52353442', 'P27', 'Q29'), ('Q52353442-P31-Q5-b46dfbd7-0', 'Q52353442', 'P31', 'Q5'), ('Q52353442-P496-8da475-0e4b3f94-0', 'Q52353442', 'P496', '"0000-0001-8465-8341"'), ('Q52353442-P549-828a1d-27c6ad2a-0', 'Q52353442', 'P549', '"50269"'), ('Q52353442-P569-5069f4-2d4c6716-0', 'Q52353442', 'P569', '^1960-01-01T00:00:00Z/7'), ('Q52353442-P69-Q190080-0c0aa4a5-0', 'Q52353442', 'P69', 'Q190080'), ('Q52353442-P69-Q25864-c80df10a-0', 'Q52353442', 'P69', 'Q25864'), ('Q52353442-P735-Q21045740-8683ff36-0', 'Q52353442', 'P735', 'Q21045740'), ('Q52353442-P7859-05a463-a41340d1-0', 'Q52353442', 'P7859', '"lccn-nb2005019080"'), ('Q52353442-P864-a1d270-2d5566dc-0', 'Q52353442', 'P864', '"81100348498"'), ('Q52353442-P949-cf102a-4130dc7e-0', 'Q52353442', 'P949', '"002253159"')]

# specifying a query in top-level command syntax (the name is not really needed, just for illustration):
>>> query = api.get_query(cmd='''--name get_node_edges_via_cmd -i claims --max-cache 100
                                 --match '(n)-[r]->(n2)'
                                 --where 'n=$NODE'
                                 --return 'r as id, n as node1, r.label as label, n2 as node2'
                              ''')
... ... ... ... 
[2021-12-03 16:18:40 query]: SQL Translation:
---------------------------------------------
  SELECT graph_1_c1."id" "_aLias.id", graph_1_c1."node1" "_aLias.node1", graph_1_c1."label" "_aLias.label", graph_1_c1."node2" "_aLias.node2"
     FROM graph_1 AS graph_1_c1
     WHERE (graph_1_c1."node1" = ?)
     LIMIT ?
  PARAS: [('NODE',), 100000]
---------------------------------------------

# parameterized queries can then be issued as before:
>>> api.execute_query(query, NODE='Q102278051')
(('Q102278051-P184-Q102109890-f8d13f77-0', 'Q102278051', 'P184', 'Q102109890'), ('Q102278051-P185-Q102408991-76042e04-0', 'Q102278051', 'P185', 'Q102408991'), ('Q102278051-P31-Q5-8e8f10f9-0', 'Q102278051', 'P31', 'Q5'), ('Q102278051-P549-f21c68-c57207a7-0', 'Q102278051', 'P549', '"91301"'), ('Q102278051-P69-Q681025-22a4731f-0', 'Q102278051', 'P69', 'Q681025'), ('Q102278051-P735-Q632842-3ffb9f39-0', 'Q102278051', 'P735', 'Q632842'))

>>> api.close()
>>> 
"""



import warnings


def deprecated(msg=''):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn("{}: this function is deprecated. {}".format(func.__name__, msg))
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator



"""
Run Kypher query engine.
"""

import sys
import os
import os.path
import tempfile
import io
import argparse

from kgtk.exceptions import KGTKException

from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles

DEFAULT_GRAPH_CACHE_FILE = os.path.join(
    tempfile.gettempdir(), 'kgtk-graph-cache-%s.sqlite3.db' % os.environ.get('USER', ''))

def parser():
    desc = ('Query one or more KGTK files with Kypher.\n' +
            'IMPORTANT: input can come from stdin but chaining queries is not yet supported.')
    return {
        'help': desc,
        'description': desc,
    }

EXPLAIN_MODES = ('plan', 'full', 'expert')
DEFAULT_INDEX_MODE = 'auto'

class InputOptionAction(argparse.Action):
    """Special-purpose argparse action that associates an input-specific option
    (such as an alias) to the most recently parsed input file.  NOTE: the 'dest'
    value will be used as the key for the specific option in 'input_file_options'.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        input_options = getattr(namespace, 'input_file_options', {}) or {}
        inputs = KGTKArgumentParser.get_input_file_list(getattr(namespace, 'input_files', []))
        if len(inputs) < 1:
            raise KGTKException('out-of-place input option: %s' % option_string)
        # normalize path objects to strings:
        input_file = str(inputs[-1])
        # handle boolean args (also requires nargs=0):
        if self.type == bool:
            values = True
        # we use self.dest as the key for this particular option:
        cur_options = input_options.get(input_file, {})
        cur_values = cur_options.get(self.dest)
        if isinstance(cur_values, list):
            # handle multiple specs of multi-valued options via append:
            values = cur_values + values
        input_options.setdefault(input_file, {})[self.dest] = values
        setattr(namespace, 'input_file_options', input_options)

class MatchOptionAction(argparse.Action):
    """Special-purpose argparse action that handles --match, --optional and --with
    and associates any --where option with the appropriate clause preceding it.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        match_options = getattr(namespace, 'match_options', []) or []
        if option_string in ('--match', '--opt', '--optional', '--with'):
            match_options.append([option_string, values, None])
            # for --match/--where/--with use the top-level option destinations:
            if option_string == '--match':
                setattr(namespace, self.dest, values)
            if option_string == '--with':
                setattr(namespace, self.dest, values)
        elif option_string == '--where':
            if len(match_options) < 1 or match_options[-1][2] is not None:
                raise KGTKException('out-of-place --where option: %s' % values)
            match_options[-1][2] = values
            if match_options[-1][0] == '--match':
                setattr(namespace, self.dest, values)
            if match_options[-1][0] == '--with':
                setattr(namespace, 'with_where', values)
        setattr(namespace, 'match_options', match_options)

def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args):
    parser.accept_shared_argument('_debug')
    parser.accept_shared_argument('_expert')

    parser.add_input_file(options=["-i", "--input-files"], dest='input_files',
                          # default_stdin is what makes it not required which is what we need for --show-cache:
                          allow_list=True, default_stdin=True, allow_stdin=True,
                          who="One or more input files to query, maybe compressed")
    parser.add_argument('--as', metavar='NAME', default={}, action=InputOptionAction, dest='alias',
                        help="alias name to be used for preceding input")
    parser.add_argument('--comment', default=None, action=InputOptionAction, dest='comment',
                        help="comment string to store for the preceding input (displayed by --show-cache)")
    # future extension:
    #parser.add_argument('--in-memory', default=False, type=bool, nargs=0, action=InputOptionAction, dest='in_memory',
    #                    help="load the preceding input into a temporary in-memory table only")
    parser.add_argument('-a', '--append', metavar='FILE', nargs='+', default=None,
                        action=InputOptionAction, dest='append',
                        help="additional data file(s) to append to the specified input")
    parser.add_argument('--query', default=None, action='store', dest='query',
                        help="complete Kypher query combining all clauses," +
                        " if supplied, all other specialized clause arguments will be ignored")
    parser.add_argument('--match', metavar='PATTERN', default='()', action=MatchOptionAction, dest='match',
                        help="MATCH pattern of a Kypher query, defaults to universal node pattern `()'")
    parser.add_argument('--where', metavar='CLAUSE', default=None, action=MatchOptionAction, dest='where',
                        help="WHERE clause to a preceding --match, --opt or --with clause")
    parser.add_argument('--opt', '--optional', metavar='PATTERN', default=None, action=MatchOptionAction,
                        help="OPTIONAL MATCH pattern(s) of a Kypher query (zero or more)")
    parser.add_argument('--with', metavar='CLAUSE', default='*', action=MatchOptionAction, dest='with',
                        help="WITH clause of a Kypher query (only 'WITH * ...' is currently supported)")
    parser.add_argument('--where:', metavar='CLAUSE', default=None, action='store', dest='with_where',
                        help="final global WHERE clause, shorthand for 'WITH * WHERE ...'")
    parser.add_argument('--return', metavar='CLAUSE', default='*', action='store', dest='return',
                        help="RETURN clause of a Kypher query (defaults to *)")
    parser.add_argument('--order-by', metavar='CLAUSE', default=None, action='store', dest='order',
                        help="ORDER BY clause of a Kypher query")
    parser.add_argument('--skip', metavar='CLAUSE', default=None, action='store', dest='skip',
                        help="SKIP clause of a Kypher query")
    parser.add_argument('--limit', metavar='CLAUSE', default=None, action='store', dest='limit',
                        help="LIMIT clause of a Kypher query")
    parser.add_argument('--multi', metavar='N', type=int, default=None, action='store', dest='multi_edge',
                        help="split each result into N separate edges or output rows of equal number of columns")
    parser.add_argument('--para', metavar='NAME=VAL', action='append', dest='regular_paras',
                        help="zero or more named value parameters to be passed to the query")
    parser.add_argument('--spara', metavar='NAME=VAL', action='append', dest='string_paras',
                        help="zero or more named string parameters to be passed to the query")
    parser.add_argument('--lqpara', metavar='NAME=VAL', action='append', dest='lqstring_paras',
                        help="zero or more named LQ-string parameters to be passed to the query")
    parser.add_argument('--no-header', action='store_true', dest='no_header',
                        help="do not generate a header row with column names")
    parser.add_argument('--force', action='store_true', dest='force',
                        help="force problematic queries to run against advice")
    parser.add_argument('--index', '--index-mode', metavar='MODE', nargs='+', action='store',
                        dest='index_mode', default=[DEFAULT_INDEX_MODE], 
                        help="default index creation MODE for all inputs"
                        + f" (default: {DEFAULT_INDEX_MODE});"
                        + " can be overridden with --idx for specific inputs")
    parser.add_argument('--idx', '--input-index', metavar='SPEC', nargs='+', default=None,
                        action=InputOptionAction, dest='index_specs',
                        help="create index(es) according to SPEC for the preceding input only")
    parser.add_argument('--explain', metavar='MODE', nargs='?', action='store', dest='explain',
                        choices=EXPLAIN_MODES, const=EXPLAIN_MODES[0], 
                        help="explain the query execution and indexing plan according to MODE"
                        + " (%(choices)s, default: %(const)s)."
                        + " This will not actually run or create anything.")
    parser.add_argument('--graph-cache', '--gc', action='store', dest='graph_cache_file',
                        help="database cache where graphs will be imported before they are queried"
                        + " (defaults to per-user temporary file)")
    parser.add_argument('--show-cache', '--sc', action='store_true', dest='show_cache',
                        help="describe the current content of the graph cache and exit"
                        + " (does not actually run a query or import data)")
    parser.add_argument('--read-only', '--ro', action='store_true', dest='readonly',
                        help="do not create or update the graph cache in any way"
                        + ", only run queries against already imported and indexed data")
    parser.add_argument('--import', metavar='MODULE_LIST', default=None, action='store', dest='import',
                        help="Python modules needed to define user extensions to built-in functions")
    parser.add_argument('-o', '--out', default='-', action='store', dest='output',
                        help="output file to write to, if `-' (the default) output goes to stdout."
                        + " Files with extensions .gz, .bz2 or .xz will be appropriately compressed.")

def import_modules():
    """Import command-specific modules that are only needed when we actually run.
    """
    mod = sys.modules[__name__]
    import sh
    setattr(mod, "sh", sh)
    import csv
    setattr(mod, "csv", csv)
    import kgtk.kypher.query as kyquery
    setattr(mod, "kyquery", kyquery)
    import kgtk.kypher.sqlstore as sqlstore
    setattr(mod, "sqlstore", sqlstore)

def parse_query_parameters(regular=[], string=[], lqstring=[]):
    """Parse and DWIM any supplied parameter values and return as a dictionary.
    """
    para_specs = {'regular': regular, 'string': string, 'lqstring': lqstring}
    parameters = {}
    for ptype in ('regular', 'string', 'lqstring'):
        for pspec in para_specs[ptype]:
            eqpos = pspec.find('=')
            if eqpos < 0:
                raise KGTKException('Illegal parameter spec: %s' % pspec)
            name = pspec[0:eqpos]
            value = pspec[eqpos+1:]
            if ptype == 'string':
                value = kyquery.dwim_to_string_para(value)
            if ptype == 'lqstring':
                value = kyquery.dwim_to_lqstring_para(value)
            parameters[name] = value
    return parameters

def preprocess_query_options(input_files: KGTKFiles, **options):
    """Preprocess and normalize query options before they are passed on.
    """
    import_modules()
    debug = options.get('_debug', False)
    expert = options.get('_expert', False)
    loglevel = debug and 1 or 0
        
    if debug and expert:
        loglevel = 2
        print('OPTIONS:', options)
    options['loglevel'] = loglevel

    # normalize input path objects to strings:
    inputs = input_files
    # do not accept the stdin default for empty inputs, but allow empty inputs for --show-cache:
    if len(inputs) > 0:
        inputs = [str(f) for f in KGTKArgumentParser.get_input_file_list(input_files)]
    options['input_files'] = inputs

    # normalize output to an open writeable stream:
    output = options.get('output')
    if output == '-':
        output = sys.stdout
    if isinstance(output, str):
        output = sqlstore.open_to_write(output, mode='wt')
    options['output'] = output

    optionals = [(pat, where) for (opt, pat, where) in options.get('match_options', [])
                 if opt in ('--opt', '--optional')]
    if optionals:
        options['optionals'] = optionals

    with_ = (options.get('with'), options.get('with_where'))
    if with_.count(None) < len(with_):
        options['with'] = with_
    
    parameters = parse_query_parameters(regular=options.get('regular_paras') or [],
                                        string=options.get('string_paras') or [],
                                        lqstring=options.get('lqstring_paras') or [])
    options['parameters'] = parameters

    imports = options.get('import')

    graph_cache = options.get('graph_cache_file')
    if graph_cache is None or len(graph_cache) == 0:
        graph_cache = os.getenv('KGTK_GRAPH_CACHE')
    if graph_cache is None or len(graph_cache) == 0:
        graph_cache = DEFAULT_GRAPH_CACHE_FILE
    options['graph_cache_file'] = graph_cache

    return options


def run(input_files: KGTKFiles, **options):
    """Run Kypher query according to the provided command-line arguments.
    """
    try:
        options = preprocess_query_options(input_files=input_files, **options)
        show_cache = options.get('show_cache')
        inputs = options.get('input_files')
        if len(inputs) == 0 and not show_cache:
            raise KGTKException('At least one input needs to be supplied')
        output = options['output']
        loglevel = options.get('loglevel')

        store = None
        try:
            graph_cache = options.get('graph_cache_file')
            store = sqlstore.SqliteStore(graph_cache, create=not os.path.exists(graph_cache),
                                         loglevel=loglevel, readonly=options.get('readonly'))

            if show_cache:
                store.describe_meta_tables(out=sys.stdout)
                return

            imports = options.get('import')
            imports and exec('import ' + imports, sqlstore.__dict__)
        
            query = kyquery.KgtkQuery(inputs, store, loglevel=loglevel,
                                      options=options.get('input_file_options'),
                                      query=options.get('query'),
                                      match=options.get('match'),
                                      where=options.get('where'),
                                      optionals=options.get('optionals'),
                                      with_=options.get('with'),
                                      ret=options.get('return'),
                                      order=options.get('order'),
                                      skip=options.get('skip'),
                                      limit=options.get('limit'),
                                      multi=options.get('multi_edge'),
                                      parameters=options.get('parameters'),
                                      index=options.get('index_mode'),
                                      force=options.get('force'))
            
            explain = options.get('explain')
            if explain is not None:
                result = query.explain(explain)
                output.write(result)
            else:
                result = query.execute()
                # we are forcing \n line endings here instead of \r\n, since those
                # can be re/imported efficiently with the new SQLite import command;
                # we force `escapechar' back to None to avoid generation of double
                # backslashes as in 'Buffalo \'66', which in turn will now raise errors
                # if separators in fields are encountered (which seems what we want):
                csvwriter = csv.writer(output, dialect=None, delimiter='\t',
                                       quoting=csv.QUOTE_NONE, quotechar=None,
                                       lineterminator='\n',
                                       escapechar=None)
                if not options.get('no_header'):
                    csvwriter.writerow(query.result_header)
                csvwriter.writerows(result)
                
            output.flush()
        finally:
            if store is not None:
                store.close()
            if output is not None and output is not sys.stdout:
                output.close()
        
    except sh.SignalException_SIGPIPE:
        # hack to work around Python3 issue when stdout is gone when we try to report an exception;
        # without this we get an ugly 'Exception ignored...' msg when we quit with head or a pager:
        sys.stdout = os.fdopen(1)
    except KGTKException as e:
        raise e
    except Exception as e:
        raise KGTKException(str(e) + '\n')



"""
KGTK math SQL functions
"""

import sys
# sqlite3 already loads math, so no extra cost:
import math
import re
from   functools import lru_cache

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *
from   kgtk.kypher.functions import SqlFunction, sqlfun


### Math functions:

# Temporary Python implementation of SQLite math built-ins until they become standardly available.
# Should happen once SQLite3 3.35.0 is used by Python - or soon thereafter.  Once we've determined
# the cutoff point we can make the function registration dependent on 'sqlite3.version'.
# User-defined functions override built-ins, which means this should work even after math built-ins
# come online - we hope.

@sqlfun(name='acos', num_params=1, deterministic=True)
def math_acos(x):
    """Implement the SQLite3 math built-in 'acos' via Python.
    """
    try:
        return math.acos(x)
    except:
        pass

@sqlfun(name='acosh', num_params=1, deterministic=True)
def math_acosh(x):
    """Implement the SQLite3 math built-in 'acosh' via Python.
    """
    try:
        return math.acosh(x)
    except:
        pass

@sqlfun(name='asin', num_params=1, deterministic=True)
def math_asin(x):
    """Implement the SQLite3 math built-in 'asin' via Python.
    """
    try:
        return math.asin(x)
    except:
        pass

@sqlfun(name='asinh', num_params=1, deterministic=True)
def math_asinh(x):
    """Implement the SQLite3 math built-in 'asinh' via Python.
    """
    try:
        return math.asinh(x)
    except:
        pass

@sqlfun(name='atan', num_params=1, deterministic=True)
def math_atan(x):
    """Implement the SQLite3 math built-in 'atan' via Python.
    """
    try:
        return math.atan(x)
    except:
        pass

@sqlfun(name='atan2', num_params=2, deterministic=True)
def math_atan2(x, y):
    """Implement the SQLite3 math built-in 'atan2' via Python.
    """
    try:
        return math.atan2(y, x) # flips args
    except:
        pass

@sqlfun(name='atanh', num_params=1, deterministic=True)
def math_atanh(x):
    """Implement the SQLite3 math built-in 'atanh' via Python.
    """
    try:
        return math.atanh(x)
    except:
        pass

@sqlfun(name='ceil', num_params=1, deterministic=True)
def math_ceil(x):
    """Implement the SQLite3 math built-in 'ceil' via Python.
    """
    try:
        return math.ceil(x)
    except:
        pass

# alias: ceiling(X)
SqlFunction(name='ceiling', code=math_ceil, num_params=1, deterministic=True).define()
    
@sqlfun(name='cos', num_params=1, deterministic=True)
def math_cos(x):
    """Implement the SQLite3 math built-in 'cos' via Python.
    """
    try:
        return math.cos(x)
    except:
        pass

@sqlfun(name='cosh', num_params=1, deterministic=True)
def math_cosh(x):
    """Implement the SQLite3 math built-in 'cosh' via Python.
    """
    try:
        return math.cosh(x)
    except:
        pass

@sqlfun(name='degrees', num_params=1, deterministic=True)
def math_degrees(x):
    """Implement the SQLite3 math built-in 'degrees' via Python.
    Convert value X from radians into degrees. 
    """
    try:
        return math.degrees(x)
    except:
        pass

@sqlfun(name='exp', num_params=1, deterministic=True)
def math_exp(x):
    """Implement the SQLite3 math built-in 'exp' via Python.
    """
    try:
        return math.exp(x)
    except:
        pass

@sqlfun(name='floor', num_params=1, deterministic=True)
def math_floor(x):
    """Implement the SQLite3 math built-in 'floor' via Python.
    """
    try:
        return math.floor(x)
    except:
        pass

# NOTE: naming and invocation of logarithm functions is different from
# standard SQL or Python math for that matter (more like Postgres).

@sqlfun(name='ln', num_params=1, deterministic=True)
def math_ln(x):
    """Implement the SQLite3 math built-in 'ln' via Python.
    """
    try:
        return math.log(x)
    except:
        pass

@sqlfun(name='log10', num_params=1, deterministic=True)
def math_log10(x):
    """Implement the SQLite3 math built-in 'log10' via Python.
    """
    try:
        return math.log10(x)
    except:
        pass

# alias: log(X)
SqlFunction(name='log', code=math_log10, num_params=1, deterministic=True).define()

# this one needs to stay if we conditionalize on availability of real math built-ins:
@sqlfun(name='logb', num_params=2, deterministic=True)
def math_logb(b, x):
    """Implement the SQLite3 math built-in 'log(b,x)' via Python.
    NOTE: this uses a different name, since we cannot support optionals
    (which would require special handling in the query translator).
    This means the function needs to stay even if we use the real built-ins.
    """
    try:
        return math.log(x, b)
    except:
        pass

@sqlfun(name='log2', num_params=1, deterministic=True)
def math_log2(x):
    """Implement the SQLite3 math built-in 'log2' via Python.
    """
    try:
        return math.log2(x)
    except:
        pass

@sqlfun(name='mod', num_params=2, deterministic=True)
def math_mod(x, y):
    """Implement the SQLite3 math built-in 'mod' via Python.
    """
    try:
        return math.fmod(x, y) # preferred over 'x % y' for floats
    except:
        pass

@sqlfun(name='pi', num_params=0, deterministic=True)
def math_pi():
    """Implement the SQLite3 math built-in 'pi' via Python.
    """
    return math.pi

@sqlfun(name='pow', num_params=2, deterministic=True)
def math_pow(x, y):
    """Implement the SQLite3 math built-in 'pow' via Python.
    """
    try:
        return math.pow(x, y)
    except:
        pass

# alias: power(X,Y)
SqlFunction(name='power', code=math_pow, num_params=2, deterministic=True).define()

@sqlfun(name='radians', num_params=1, deterministic=True)
def math_radians(x):
    """Implement the SQLite3 math built-in 'radians' via Python.
    """
    try:
        return math.radians(x)
    except:
        pass

@sqlfun(name='sin', num_params=1, deterministic=True)
def math_sin(x):
    """Implement the SQLite3 math built-in 'sin' via Python.
    """
    try:
        return math.sin(x)
    except:
        pass

@sqlfun(name='sinh', num_params=1, deterministic=True)
def math_sinh(x):
    """Implement the SQLite3 math built-in 'sinh' via Python.
    """
    try:
        return math.sinh(x)
    except:
        pass

@sqlfun(name='sqrt', num_params=1, deterministic=True)
def math_sqrt(x):
    """Implement the SQLite3 math built-in 'sqrt' via Python.
    """
    try:
        return math.sqrt(x)
    except:
        pass

@sqlfun(name='tan', num_params=1, deterministic=True)
def math_tan(x):
    """Implement the SQLite3 math built-in 'tan' via Python.
    """
    try:
        return math.tan(x)
    except:
        pass

@sqlfun(name='tanh', num_params=1, deterministic=True)
def math_tanh(x):
    """Implement the SQLite3 math built-in 'tanh' via Python.
    """
    try:
        return math.tanh(x)
    except:
        pass

@sqlfun(name='trunc', num_params=1, deterministic=True)
def math_trunc(x):
    """Implement the SQLite3 math built-in 'trunc' via Python.
    """
    try:
        return math.trunc(x)
    except:
        pass



"""Convert edge file and optional node file to html visualization
"""
from argparse import Namespace
from kgtk.cli_argparse import KGTKArgumentParser, KGTKFiles
from kgtk.io.kgtkreader import KgtkReader, KgtkReaderOptions
from kgtk.value.kgtkvalueoptions import KgtkValueOptions
from kgtk.utils.color_styles import color_style_dict, CATEGORICAL, GRADIENT

gradient_color_styles = ', '.join([k for k, v in color_style_dict.items() if v['style'] == GRADIENT])
categorical_color_styles = ', '.join([k for k, v in color_style_dict.items() if v['style'] == CATEGORICAL])


def parser():
    return {
        'help': 'Convert edge file to html visualization',
        'description': 'Convert edge file (optional node file)' +
                       ' to html graph visualization file'
    }


def add_arguments_extended(parser: KGTKArgumentParser, parsed_shared_args: Namespace):
    """
    Parse arguments
    Args:
        parser (argparse.ArgumentParser)
    """

    _expert: bool = parsed_shared_args._expert

    # This helper function makes it easy to suppress options from
    # The help message.  The options are still there, and initialize
    # what they need to initialize.

    parser.add_input_file(positional=True)
    parser.add_output_file()

    parser.add_argument('--node-file',
                        dest='node_file',
                        type=str,
                        default=None,
                        help="Path of the node file.")

    parser.add_argument('--direction',
                        dest='direction',
                        type=str,
                        default=None,
                        help="The edge direction: arrow|particle|None. Default: None")

    parser.add_argument('--show-edge-label',
                        dest='edge_label',
                        action='store_true',
                        default=False,
                        help="Add this option to show labels on edges. Default: False")

    parser.add_argument('--edge-color-column',
                        dest='edge_color_column',
                        type=str,
                        default=None,
                        help="Column for edge colors in the edge file. "
                             "The values can be numbers, hex codes or any strings")

    parser.add_argument('--edge-color-numbers',
                        dest='edge_color_numbers',
                        type=str,
                        default=None,
                        help="Indicate if the values in the --edge-color-column are numbers and specify how to "
                             "scale the numbers. The valid choices are:linear|log|as-is. "
                             "Default: None. "
                             "The numbers will be scaled linearly, logarithmically, left as is respectively.")

    parser.add_argument('--edge-color-hex',
                        dest='edge_color_hex',
                        action='store_true',
                        default=False,
                        help="Add this option if the values in the --edge-color-column are valid hexadecimal colors."
                             "Valid hexadecimal colors start with # and are of 3 or 6 length (without the #) ")

    parser.add_argument('--edge-color-style',
                        dest='edge_color_style',
                        type=str,
                        default='d3.interpolateRainbow',
                        help=f"Pick one of the following CATEGORICAL styles: \n{categorical_color_styles}\n"
                             f"OR one of the following GRADIENT styles:  \n{gradient_color_styles}\n"
                             f"Default: d3.interpolateRainbow."
                             f"See https://github.com/d3/d3-scale-chromatic for more details.")

    parser.add_argument('--edge-color-default',
                        dest='edge_color_default',
                        type=str,
                        default='#000000',
                        help="Default color for edges. Default: '#000000'")

    parser.add_argument('--edge-width-column',
                        dest='edge_width_column',
                        type=str,
                        default=None,
                        help="Column for edge widths in the edge file. The values should be numbers.")

    parser.add_argument('--edge-width-minimum',
                        dest='edge_width_minimum',
                        type=float,
                        default=1.0,
                        help="Minimum edge width. Default: 1.0")

    parser.add_argument('--edge-width-maximum',
                        dest='edge_width_maximum',
                        type=float,
                        default=5.0,
                        help="Maximum edge width. Default: 5.0")

    parser.add_argument('--edge-width-default',
                        dest='edge_width_default',
                        type=float,
                        default=1.0,
                        help="Default edge width. Default: 1.0")

    parser.add_argument('--edge-width-scale',
                        dest='edge_width_scale',
                        type=str,
                        default=None,
                        help="Edge width scale: linear|log. Default: None")

    parser.add_argument('--node-color-column',
                        dest='node_color_column',
                        type=str,
                        default=None,
                        help="Column for node colors in the --node-file. The values can be numbers, valid hex codes"
                             " or any strings.")

    parser.add_argument('--node-color-style',
                        dest='node_color_style',
                        type=str,
                        default='d3.interpolateRainbow',
                        help=f"Pick one of the following CATEGORICAL styles: \n{categorical_color_styles}\n"
                             f"OR one of the following GRADIENT styles:  \n{gradient_color_styles}\n"
                             f"Default: d3.interpolateRainbow."
                             f"See https://github.com/d3/d3-scale-chromatic for more details.")

    parser.add_argument('--node-color-default',
                        dest='node_color_default',
                        type=str,
                        default='#000000',
                        help="Default node color. Default: '#000000'")

    parser.add_argument('--node-color-numbers',
                        dest='node_color_numbers',
                        type=str,
                        default=None,
                        help="Indicate if the values in the --node-color-column are numbers and specify how to "
                             "scale the numbers. The valid choices are:linear|log|as-is. "
                             "Default: None. "
                             "The numbers will be scaled linearly, logarithmically, left as is respectively.")

    parser.add_argument('--node-color-hex',
                        dest='node_color_hex',
                        action='store_true',
                        default=False,
                        help="Add this option if the values in the --node-color-column are valid hexadecimal colors."
                             "Valid hexadecimal colors start with # and are of 3 or 6 length (without the #) ")

    parser.add_argument('--node-size-column',
                        dest='node_size_column',
                        type=str,
                        default=None,
                        help="Column for node sizes in the --node-file. Default: None")

    parser.add_argument('--node-size-minimum',
                        dest='node_size_minimum',
                        type=float,
                        default=1.0,
                        help="Minimum node size. Default: 1.0")

    parser.add_argument('--node-size-maximum',
                        dest='node_size_maximum',
                        type=float,
                        default=5.0,
                        help="Maximum node size. Default: 5.0")

    parser.add_argument('--node-size-default',
                        dest='node_size_default',
                        type=float,
                        default=2.0,
                        help="Default node size. Default: 2.0")

    parser.add_argument('--node-size-scale',
                        dest='node_size_scale',
                        type=str,
                        default=None,
                        help="Node size scale: linear|log. Default: None")

    parser.add_argument('--node-file-id',
                        dest='node_file_id',
                        type=str,
                        default='id',
                        help="ID column name in the --node-file. Default: 'id'")

    parser.add_argument('--show-text-limit',
                        dest='show_text_limit',
                        type=int,
                        default=500,
                        help="When number of nodes is greater than --show-text-limit, node labels will not be visible."
                             "Default: 500")

    parser.add_argument('--node-border-color',
                        dest='node_border_color',
                        type=str,
                        default=None,
                        help="Node border color. Default: None")

    parser.add_argument('--tooltip-column',
                        dest='tooltip_column',
                        type=str,
                        default=None,
                        help="Column for node tooltips in the --node-file. Default: None")

    parser.add_argument('--show-text',
                        dest='show_text',
                        type=str,
                        default=None,
                        help="Show node labels at the position relative to node: center|above. Default: None. "
                             "If the number of nodes in the graph is greater than specified by "
                             "--show-text-limit option, which is 500 by default, "
                             "then the text will not be shown in the visualization.")

    parser.add_argument('--show-blank-labels',
                        dest='show_blank_labels',
                        action='store_true',
                        default=False,
                        help="if --show-text is specified, show the label of a Qnode as emptry string, and not the "
                             "Qnode, if the label is an empty string.")

    KgtkReader.add_debug_arguments(parser, expert=_expert)
    KgtkReaderOptions.add_arguments(parser, mode_options=True, expert=_expert)
    KgtkValueOptions.add_arguments(parser, expert=_expert)


def run(input_file: KGTKFiles,
        output_file: KGTKFiles,
        errors_to_stdout: bool = False,
        errors_to_stderr: bool = True,
        show_options: bool = False,
        verbose: bool = False,
        very_verbose: bool = False,
        node_file: str = None,
        direction: str = None,
        edge_label: bool = False,
        edge_color_column: str = None,
        edge_color_hex: bool = False,
        edge_color_numbers: str = None,
        edge_color_style: str = 'd3.interpolateRainbow',
        edge_color_default: str = '#000000',
        edge_width_column: str = None,
        edge_width_default: float = 1.0,
        edge_width_minimum: float = 1.0,
        edge_width_maximum: float = 5.0,
        edge_width_scale: str = None,
        node_color_column: str = None,
        node_color_numbers: str = None,
        node_color_hex: bool = False,
        node_color_style: str = 'd3.interpolateRainbow',
        node_color_default: str = '#000000',
        node_size_column: str = None,
        node_size_default: float = 2.0,
        node_size_minimum: float = 1.0,
        node_size_maximum: float = 5.0,
        node_size_scale: str = None,
        node_file_id: str = 'id',
        show_text_limit: int = 500,
        node_border_color: str = None,
        tooltip_column: str = None,
        show_text: str = None,
        show_blank_labels: bool = False,
        **kwargs  # Whatever KgtkFileOptions and KgtkValueOptions want.
        ) -> int:
    from kgtk.exceptions import KGTKException
    if node_color_style not in color_style_dict:
        raise KGTKException(f'Unknown value: {node_color_style} for option: --node-color-style. Please choose'
                            f' one of the following: {", ".join(list(color_style_dict))}')

    if edge_color_style not in color_style_dict:
        raise KGTKException(f'Unknown value: {edge_color_style} for option: --edge-color-style. Please choose'
                            f' one of the following: {", ".join(list(color_style_dict))}')

    from kgtk.visualize.visualize_api import KgtkVisualize
    kv: KgtkVisualize = KgtkVisualize(
        input_file=input_file,
        output_file=output_file,
        errors_to_stdout=errors_to_stdout,
        errors_to_stderr=errors_to_stderr,
        show_options=show_options,
        verbose=verbose,
        very_verbose=very_verbose,
        node_file=node_file,
        direction=direction,
        edge_label=edge_label,
        edge_color_column=edge_color_column,
        edge_color_hex=edge_color_hex,
        edge_color_numbers=edge_color_numbers,
        edge_color_style=edge_color_style,
        edge_color_default=edge_color_default,
        edge_width_column=edge_width_column,
        edge_width_default=edge_width_default,
        edge_width_minimum=edge_width_minimum,
        edge_width_maximum=edge_width_maximum,
        edge_width_scale=edge_width_scale,
        node_color_column=node_color_column,
        node_color_hex=node_color_hex,
        node_color_numbers=node_color_numbers,
        node_color_style=node_color_style,
        node_color_default=node_color_default,
        node_size_column=node_size_column,
        node_size_default=node_size_default,
        node_size_minimum=node_size_minimum,
        node_size_maximum=node_size_maximum,
        node_size_scale=node_size_scale,
        node_file_id=node_file_id,
        show_text_limit=show_text_limit,
        node_border_color=node_border_color,
        tooltip_column=tooltip_column,
        show_text=show_text,
        show_blank_labels=show_blank_labels,
        kwargs=kwargs
    )
    kv.execute()



"""
Obsolete info table API for access to file and graph info tables.
"""

from   functools import lru_cache

from   kgtk.exceptions import KGTKException
from   kgtk.kypher.utils import *


FILE_TABLE = sdict[
    '_name_': 'fileinfo',
    'columns': sdict[
        'file':    sdict['_name_': 'file',    'type': 'TEXT', 'key': True, 'doc': 'real path of the file containing the data'],
        'size':    sdict['_name_': 'size',    'type': 'INTEGER'],
        'modtime': sdict['_name_': 'modtime', 'type': 'FLOAT'],
        'md5sum':  sdict['_name_': 'md5sum',  'type': 'TEXT', 'default': None], # just for illustration of defaults
        'graph':   sdict['_name_': 'graph',   'type': 'TEXT', 'doc': 'the graph defined by the data of this file'],
        'comment': sdict['_name_': 'comment', 'type': 'TEXT', 'doc': 'comment describing the data of this file'],
    ],
]

GRAPH_TABLE = sdict[
    '_name_': 'graphinfo',
    'columns': sdict[
        'name':    sdict['_name_': 'name',    'type': 'TEXT', 'key': True, 'doc': 'name of the table representing this graph'],
        'shasum':  sdict['_name_': 'shasum',  'type': 'TEXT', 'doc': 'table hash computed by sqlite shasum command'],
        'header':  sdict['_name_': 'header',  'type': 'TEXT'],
        'size':    sdict['_name_': 'size',    'type': 'INTEGER', 'doc': 'total size in bytes used by this graph including indexes'],
        'acctime': sdict['_name_': 'acctime', 'type': 'FLOAT', 'doc': 'last time this graph was accessed'],
        'indexes': sdict['_name_': 'indexes', 'type': 'TEXT',  'doc': 'list of sdicts for indexes defined on this graph'],
    ],
    'without_rowid': False,
    'temporary': False,
]


class InfoTable(object):
    """Obsolete info table API for access to file and graph info tables.
    """

    def __init__(self, store, schema):
        """Create an info table object for 'schema' stored in 'store'.
        """
        self.store = store
        self.schema = schema
        self.verified_schema = False

    def init_table(self):
        """If the info table doesn't exist yet, define it from its schema.
        """
        if not self.store.has_table(self.schema._name_):
            self.store.execute(self.store.get_table_definition(self.schema))
            self.verified_schema = True

    def clear_caches(self):
        InfoTable.get_info.cache_clear()
        InfoTable.get_all_keys.cache_clear()
        InfoTable.get_all_infos.cache_clear()

    @lru_cache(maxsize=None)
    def get_info(self, key):
        """Return a dict info structure for the row identified by 'key' in this info table,
        or None if 'key' does not exist.  All column keys will be set, but some values may be None.
        """
        if not self.verified_schema:
            self.handle_schema_update()
        table = self.schema._name_
        cols = self.schema.columns
        keycol = self.store.get_key_column(self.schema)
        query = 'SELECT %s FROM %s WHERE %s=?' % (self.store.get_full_column_list(self.schema), table, cols[keycol]._name_)
        for row in self.store.execute(query, (key,)):
            result = sdict()
            for col, val in zip(cols.keys(), row):
                result[col] = val
            return result
        return None

    def set_info(self, key, info):
        """Insert or update this info table for 'key' based on the values in 'info'.
        If a record based on 'key' already exists, update it, otherwise insert a new one.
        If a new record is inserted, any missing column values in 'info' will be set to None.
        If 'info' contains a value for the key column, that value will override 'key' which
        allows for an existing key value to be updated to a new one.
        """
        if self.get_info(key) is not None:
            self.update_info(key, info)
        else:
            # this is not really needed, since 'get_info' already checks:
            #if not self.verified_schema:
            #    self.handle_schema_update()
            table = self.schema._name_
            cols = self.schema.columns
            keycol = self.store.get_key_column(self.schema)
            key = info.get(keycol) or key
            info[keycol] = key
            columns = [cols[name] for name in info.keys()]
            collist = self.store.get_column_list(*columns)
            vallist = ','.join(['?'] * len(columns))
            stmt = 'INSERT INTO %s (%s) VALUES (%s)' % (table, collist, vallist)
            self.store.execute(stmt, list(info.values()))
            self.store.commit()
            self.clear_caches()

    def update_info(self, key, info):
        """Update an existing record in this info table for 'key' and the values in 'info'.
        Any column values undefined in 'info' will remain unaffected.
        If 'info' contains a value for the key column, that value will override 'key' which
        allows for an existing key value to be updated to a new one.
        This is a no-op if no record with 'key' exists in table 'schema'.
        """
        if not self.verified_schema:
            self.handle_schema_update()
        table = self.schema._name_
        cols = self.schema.columns
        keycol = self.store.get_key_column(self.schema)
        columns = [cols[name] for name in info.keys()]
        collist = self.store.get_column_list(*columns)
        collist = collist.replace(', ', '=?, ')
        stmt = 'UPDATE %s SET %s=? WHERE %s=?' % (table, collist, keycol)
        values = list(info.values())
        values.append(key)
        self.store.execute(stmt, values)
        self.store.commit()
        self.clear_caches()
        
    def drop_info(self, key):
        """Delete any rows identified by 'key' in this info table.
        """
        if not self.verified_schema:
            self.handle_schema_update()
        table = self.schema._name_
        cols = self.schema.columns
        keycol = self.store.get_key_column(self.schema)
        stmt = 'DELETE FROM %s WHERE %s=?' % (table, cols[keycol]._name_)
        self.store.execute(stmt, (key,))
        self.store.commit()
        self.clear_caches()

    @lru_cache(maxsize=None)
    def get_all_keys(self):
        table = self.schema._name_
        cols = self.schema.columns
        keycol = self.store.get_key_column(self.schema)
        return [key for (key,) in self.store.execute('SELECT %s FROM %s' % (keycol, table))]

    @lru_cache(maxsize=None)
    def get_all_infos(self):
        # TO DO: this generates one query per key, generalize if this becomes a performance issue
        return [self.get_info(key) for key in self.get_all_keys()]
    
    def handle_schema_update(self):
        """Check whether the schema of the info table on disk is compatible with this schema.
        If not, try to upgrade it by adding any new missing columns.  If the schema on disk is
        from a newer version of KGTK, raise an error.  This assumes that updates to info table
        schemas always only add new columns.  No other schema changes are supported.
        """
        if self.verified_schema:
            return
        table = self.schema._name_
        cols = self.schema.columns
        current_col_names = self.store.get_table_header(table)
        if len(current_col_names) == len(cols):
            self.verified_schema = True
            return
        if len(current_col_names) > len(cols):
            raise KGTKException('incompatible graph cache schema, please upgrade KGTK or delete the cache')
            
        try:
            col_names = [col._name_ for col in cols.values()]
            for cname in col_names:
                if cname not in current_col_names:
                    newcol = cols[cname]
                    stmt = 'ALTER TABLE %s ADD COLUMN %s %s' % (table, cname, newcol.type)
                    self.store.execute(stmt)
            self.verified_schema = True
        except Exception as e:
            raise KGTKException('sorry, error during schema upgrade, please remove graph cache and retry ( %s )' % repr(e))
